from __future__ import annotations

from collections.abc import Iterable, Iterator
from types import EllipsisType, GenericAlias, ModuleType, NotImplementedType
from collections.abc import Callable

import numpy as np

from scipy._lib._array_api import (
    array_namespace,
    is_numpy,
    ArrayLike,
    Array,
    xp_capabilities,
)
from scipy.spatial.transform import Rotation
from scipy.spatial.transform._rotation import _promote
import scipy.spatial.transform._rigid_transform_cy as cython_backend
import scipy.spatial.transform._rigid_transform_xp as xp_backend
import scipy._external.array_api_extra as xpx
from scipy._external.array_api_compat import device
from scipy._lib._array_api import xp_promote
from scipy._lib._util import broadcastable


__all__ = ["RigidTransform"]

backend_registry = {array_namespace(np.empty(0)): cython_backend}


def select_backend(xp: ModuleType, cython_compatible: bool):
    """Select the backend for the given array library.

    We need this selection function because the Cython backend for numpy does not
    support quaternions of arbitrary dimensions. We therefore only use the Array API
    backend for numpy if we are dealing with rotations of more than one leading
    dimension.
    """
    if is_numpy(xp) and not cython_compatible:
        return xp_backend
    return backend_registry.get(xp, xp_backend)


@xp_capabilities()
def normalize_dual_quaternion(dual_quat: ArrayLike) -> Array:
    """Normalize dual quaternion."""
    xp = array_namespace(dual_quat)
    dual_quat = _promote(dual_quat, xp=xp)
    single = dual_quat.ndim == 1 and is_numpy(xp)
    if single:
        dual_quat = xpx.atleast_nd(dual_quat, ndim=2, xp=xp)
    cython_compatible = dual_quat.ndim < 3
    dq = select_backend(xp, cython_compatible).normalize_dual_quaternion(dual_quat)
    if single:
        return dq[0]
    return dq


rigid_transform_extra_note = (
    """The method ``apply`` is not supported with cupy<14.*.

    """)


@xp_capabilities(
    skip_backends=[("dask.array", "missing linalg.cross/det functions")],
    method_capabilities={
        "__init__": dict(
            jax_jit=False,
            skip_backends=[("dask.array", "missing linalg.cross/det functions")],
        ),
        "from_matrix": dict(
            jax_jit=False,
            skip_backends=[("dask.array", "missing linalg.cross/det functions")],
        ),
        "__getitem__": dict(
            jax_jit=False,
            skip_backends=[
                ("dask.array", "cannot handle zero-length rigid transforms")
            ],
        ),
        "__setitem__": dict(
            jax_jit=False,
            skip_backends=[
                ("dask.array", "cannot handle zero-length rigid transforms")
            ],
        ),
        "apply": dict(
            skip_backends=[
                ("dask.array", "missing linalg.cross/det functions"),
                ("cupy", "missing .mT attribute in cupy<14.*"),
            ],
        ),
    },
    extra_note=rigid_transform_extra_note,
)
class RigidTransform:
    """
    Rigid transform in 3 dimensions.

    This class provides an interface to initialize from and represent rigid
    transforms (rotation and translation) in 3D space. In different fields,
    this type of transform may be referred to as "*pose*" (especially in
    robotics), "*extrinsic parameters*", or the "*model matrix*" (especially in
    computer graphics), but the core concept is the same: a rotation and
    translation describing the orientation of one 3D coordinate frame relative
    to another. Mathematically, these transforms belong to the Special
    Euclidean group SE(3), which encodes rotation (SO(3)) plus translation.

    The following operations on rigid transforms are supported:

    - Application on vectors
    - Transformation composition
    - Transformation inversion
    - Transformation indexing

    Note that coordinate systems must be right-handed. Because of this, this
    class more precisely represents *proper* rigid transformations in SE(3)
    rather than rigid transforms in E(3) more generally [1]_.

    Indexing within a transform is supported since multiple transforms can be
    stored within a single `RigidTransform` instance.

    Multiple transforms can be stored in a single `RigidTransform` object, which can be
    initialized using N-dimensional arrays and supports broadcasting for all
    operations.

    To create `RigidTransform` objects use ``from_...`` methods (see examples
    below). ``RigidTransform(...)`` is not supposed to be instantiated directly.

    For rigorous introductions to rigid transforms, see [2]_, [3]_, and [4]_.

    Parameters
    ----------
    matrix : array_like, shape (..., 4, 4)
        A single transformation matrix or a stack of transformation
        matrices.
    normalize : bool, optional
        If True, orthonormalize the rotation matrix using singular value
        decomposition. If False, the rotation matrix is not checked for
        orthogonality or right-handedness.
    copy : bool, optional
        If True, copy the input matrix. If False, a reference to the input
        matrix is used. If normalize is True, the input matrix is always
        copied regardless of the value of copy.

    Attributes
    ----------
    single
    rotation
    translation

    Methods
    -------
    __len__
    __getitem__
    __mul__
    __pow__
    from_matrix
    from_rotation
    from_translation
    from_components
    from_exp_coords
    from_dual_quat
    as_matrix
    as_components
    as_exp_coords
    as_dual_quat
    concatenate
    mean
    apply
    inv
    identity

    Notes
    -----
    .. versionadded:: 1.16.0

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Rigid_transformation
    .. [2] https://motion.cs.illinois.edu/RoboticSystems/CoordinateTransformations.html
    .. [3] https://www.brainvoyager.com/bv/doc/UsersGuide/CoordsAndTransforms/SpatialTransformationMatrices.html
    .. [4] Kevin M. Lynch and Frank C. Park, "Modern Robotics: Mechanics,
           Planning, and Control" Chapter 3.3, 2017, Cambridge University Press.
           https://hades.mech.northwestern.edu/images/2/25/MR-v2.pdf#page=107.31
    .. [5] Paul Furgale, "Representing Robot Pose: The good, the bad, and the
           ugly", June 9, 2014.
           https://rpg.ifi.uzh.ch/docs/teaching/2024/FurgaleTutorial.pdf

    Examples
    --------
    A `RigidTransform` instance can be initialized in any of the above formats
    and converted to any of the others. The underlying object is independent of
    the representation used for initialization.

    **Notation Conventions and Composition**

    The notation here largely follows the convention defined in [5]_. When we
    name transforms, we read the subscripts from right to left. So ``tf_A_B``
    represents a transform A <- B and can be interpreted as:

    - the coordinates and orientation of B relative to A
    - the transformation of points from B to A
    - the pose of B described in A's coordinate system

    .. parsed-literal::
        :class: highlight-none

        tf_A_B
           ^ ^
           | |
           | --- from B
           |
           ----- to A

    When composing transforms, the order is important. Transforms are not
    commutative, so in general ``tf_A_B * tf_B_C`` is not the same as
    ``tf_B_C * tf_A_B``. Transforms are composed and applied to vectors
    right-to-left. So ``(tf_A_B * tf_B_C).apply(p_C)`` is the same as
    ``tf_A_B.apply(tf_B_C.apply(p_C))``.

    When composed, transforms should be ordered such that the multiplication
    operator is surrounded by a single frame, so the frame "cancels out" and
    the outside frames are left. In the example below, B cancels out and the
    outside frames A and C are left. Or to put it another way, A <- C is the
    same as A <- B <- C.

    .. parsed-literal::
        :class: highlight-none

                      ----------- B cancels out
                      |      |
                      v      v
        tf_A_C = tf_A_B * tf_B_C
                    ^          ^
                    |          |
                    ------------ to A, from C are left

    When we notate vectors, we write the subscript of the frame that the vector
    is defined in. So ``p_B`` means the point ``p`` defined in frame B. To
    transform this point from frame B to coordinates in frame A, we apply the
    transform ``tf_A_B`` to the vector, lining things up such that the notated
    B frames are next to each other and "cancel out".

    .. parsed-literal::
        :class: highlight-none

                   ------------ B cancels out
                   |         |
                   v         v
        p_A = tf_A_B.apply(p_B)
                 ^
                 |
                 -------------- A is left

    **Visualization**

    >>> from scipy.spatial.transform import RigidTransform as Tf
    >>> from scipy.spatial.transform import Rotation as R
    >>> import numpy as np

    The following function can be used to plot transforms with Matplotlib
    by showing how they transform the standard x, y, z coordinate axes:

    >>> import matplotlib.pyplot as plt
    >>> colors = ("#FF6666", "#005533", "#1199EE")  # Colorblind-safe RGB
    >>> def plot_transformed_axes(ax, tf, name=None, scale=1):
    ...     r = tf.rotation
    ...     t = tf.translation
    ...     loc = np.array([t, t])
    ...     for i, (axis, c) in enumerate(zip((ax.xaxis, ax.yaxis, ax.zaxis),
    ...                                       colors)):
    ...         axlabel = axis.axis_name
    ...         axis.set_label_text(axlabel)
    ...         axis.label.set_color(c)
    ...         axis.line.set_color(c)
    ...         axis.set_tick_params(colors=c)
    ...         line = np.zeros((2, 3))
    ...         line[1, i] = scale
    ...         line_rot = r.apply(line)
    ...         line_plot = line_rot + loc
    ...         ax.plot(line_plot[:, 0], line_plot[:, 1], line_plot[:, 2], c)
    ...         text_loc = line[1]*1.2
    ...         text_loc_rot = r.apply(text_loc)
    ...         text_plot = text_loc_rot + t
    ...         ax.text(*text_plot, axlabel.upper(), color=c,
    ...                 va="center", ha="center")
    ...     ax.text(*tf.translation, name, color="k", va="center", ha="center",
    ...             bbox={"fc": "w", "alpha": 0.8, "boxstyle": "circle"})

    **Defining Frames**

    Let's work through an example.

    First, define the "world frame" A, also called the "base frame".
    All frames are the identity transform from their own perspective.

    >>> tf_A = Tf.identity()

    We will visualize a new frame B in A's coordinate system. So we need to
    define the transform that converts coordinates from frame B to frame A
    (A <- B).

    Physically, let's imagine constructing B from A by:

    1) Rotating A by +90 degrees around its x-axis.
    2) Translating the rotated frame 2 units in A's -x direction.

    From A's perspective, B is at [-2, 0, 0] and rotated +90 degrees about the
    x-axis, which is exactly the transform A <- B.

    >>> t_A_B = np.array([-2, 0, 0])
    >>> r_A_B = R.from_euler('xyz', [90, 0, 0], degrees=True)
    >>> tf_A_B = Tf.from_components(t_A_B, r_A_B)

    Let's plot these frames.

    >>> fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
    >>> plot_transformed_axes(ax, tf_A, name="tfA")     # A plotted in A
    >>> plot_transformed_axes(ax, tf_A_B, name="tfAB")  # B plotted in A
    >>> ax.set_title("A, B frames with respect to A")
    >>> ax.set_aspect("equal")
    >>> ax.figure.set_size_inches(6, 5)
    >>> plt.show()

    Now let's visualize a new frame C in B's coordinate system.
    Let's imagine constructing C from B by:

    1) Translating B by 2 units in its +z direction.
    2) Rotating B by +30 degrees around its z-axis.

    >>> t_B_C = np.array([0, 0, 2])
    >>> r_B_C = R.from_euler('xyz', [0, 0, 30], degrees=True)
    >>> tf_B_C = Tf.from_components(t_B_C, r_B_C)

    In order to plot these frames from a consistent perspective, we need to
    calculate the transform between A and C. Note that we do not make this
    transform directly, but instead compose intermediate transforms that let us
    get from C to A:

    >>> tf_A_C = tf_A_B * tf_B_C  # A <- B <- C

    Now we can plot these three frames from A's perspective.

    >>> fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
    >>> plot_transformed_axes(ax, tf_A, name="tfA")     # A plotted in A
    >>> plot_transformed_axes(ax, tf_A_B, name="tfAB")  # B plotted in A
    >>> plot_transformed_axes(ax, tf_A_C, name="tfAC")  # C plotted in A
    >>> ax.set_title("A, B, C frames with respect to A")
    >>> ax.set_aspect("equal")
    >>> ax.figure.set_size_inches(6, 5)
    >>> plt.show()

    **Transforming Vectors**

    Let's transform a vector from A, to B and C. To do this, we will first
    invert the transforms we already have from B and C, to A.

    >>> tf_B_A = tf_A_B.inv()  # B <- A
    >>> tf_C_A = tf_A_C.inv()  # C <- A

    Now we can define a point in A and use the above transforms to get its
    coordinates in B and C:

    >>> p1_A = np.array([1, 0, 0])  # +1 in x_A direction
    >>> p1_B = tf_B_A.apply(p1_A)
    >>> p1_C = tf_C_A.apply(p1_A)
    >>> print(p1_A)  # Original point 1 in A
    [1 0 0]
    >>> print(p1_B)  # Point 1 in B
    [3. 0. 0.]
    >>> print(p1_C)  # Point 1 in C
    [ 2.59807621 -1.5       -2.        ]

    We can also do the reverse. We define a point in C and transform it to A:

    >>> p2_C = np.array([0, 1, 0])  # +1 in y_C direction
    >>> p2_A = tf_A_C.apply(p2_C)
    >>> print(p2_C)  # Original point 2 in C
    [0 1 0]
    >>> print(p2_A)  # Point 2 in A
    [-2.5       -2.         0.8660254]

    Plot the frames with respect to A again, but also plot these two points:

    >>> fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
    >>> plot_transformed_axes(ax, tf_A, name="tfA")     # A plotted in A
    >>> plot_transformed_axes(ax, tf_A_B, name="tfAB")  # B plotted in A
    >>> plot_transformed_axes(ax, tf_A_C, name="tfAC")  # C plotted in A
    >>> ax.scatter(p1_A[0], p1_A[1], p1_A[2], color=colors[0])  # +1 x_A
    >>> ax.scatter(p2_A[0], p2_A[1], p2_A[2], color=colors[1])  # +1 y_C
    >>> ax.set_title("A, B, C frames and points with respect to A")
    >>> ax.set_aspect("equal")
    >>> ax.figure.set_size_inches(6, 5)
    >>> plt.show()

    **Switching Base Frames**

    Up to this point, we have been visualizing frames from A's perspective.
    Let's use the transforms we defined to visualize the frames from C's
    perspective.

    Now C is the "base frame" or "world frame". All frames are the identity
    transform from their own perspective.

    >>> tf_C = Tf.identity()

    We've already defined the transform C <- A, and can obtain C <- B by
    inverting the existing transform B <- C.

    >>> tf_C_B = tf_B_C.inv()  # C <- B

    This lets us plot everything from C's perspective:

    >>> fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
    >>> plot_transformed_axes(ax, tf_C, name="tfC")     # C plotted in C
    >>> plot_transformed_axes(ax, tf_C_B, name="tfCB")  # B plotted in C
    >>> plot_transformed_axes(ax, tf_C_A, name="tfCA")  # A plotted in C
    >>> ax.scatter(p1_C[0], p1_C[1], p1_C[2], color=colors[0])
    >>> ax.scatter(p2_C[0], p2_C[1], p2_C[2], color=colors[1])
    >>> ax.set_title("A, B, C frames and points with respect to C")
    >>> ax.set_aspect("equal")
    >>> ax.figure.set_size_inches(6, 5)
    >>> plt.show()
    """

    # generic type compatibility with scipy-stubs
    __class_getitem__: classmethod = classmethod(GenericAlias)

    def __init__(self, matrix: ArrayLike, normalize: bool = True, copy: bool = True):
        """Initialize from a 4x4 transformation matrix.

        Rotations are not meant to be initialized directly. Please use one of
        the `from_...` methods instead.

        Parameters
        ----------
        matrix : array_like, shape (..., 4, 4)
            A single transformation matrix or a stack of transformation
            matrices.
        normalize : bool, optional
            If True, orthonormalize the rotation matrix using singular value
            decomposition. If False, the rotation matrix is not checked for
            orthogonality or right-handedness.
        copy : bool, optional
            If True, copy the input matrix. If False, a reference to the input
            matrix is used. If normalize is True, the input matrix is always
            copied regardless of the value of copy.

        Returns
        -------
        transform : `RigidTransform` instance
        """
        xp = array_namespace(matrix)
        self._xp = xp
        matrix = _promote(matrix, xp=xp)
        if matrix.shape[-2:] != (4, 4):
            raise ValueError(
                f"Expected `matrix` to have shape (..., 4, 4), got {matrix.shape}."
            )
        # We only need the _single flag for the cython backend. The Array API backend
        # uses broadcasting by default and hence returns the correct shape without
        # additional _single logic
        self._single = matrix.ndim == 2 and is_numpy(xp)
        if self._single:
            matrix = xpx.atleast_nd(matrix, ndim=3, xp=xp)

        self._backend = select_backend(xp, matrix.ndim < 4)
        self._matrix = self._backend.from_matrix(matrix, normalize, copy)

    def __repr__(self):
        m = f"{self.as_matrix()!r}".splitlines()
        # bump indent (+27 characters)
        m[1:] = [" " * 27 + m[i] for i in range(1, len(m))]
        return "RigidTransform.from_matrix(" + "\n".join(m) + ")"

    @staticmethod
    def from_matrix(matrix: ArrayLike) -> RigidTransform:
        """Initialize from a 4x4 transformation matrix.

        Parameters
        ----------
        matrix : array_like, shape (..., 4, 4)
            Transformation matrices. Each matrix[..., :, :] represents a 4x4
            transformation matrix.

        Returns
        -------
        transform : `RigidTransform` instance
            Object containing the rigid transform represented by the input matrix.

        Notes
        -----
        4x4 rigid transformation matrices are of the form::

            [       tx]
            [   R   ty]
            [       tz]
            [ 0 0 0  1]

        where ``R`` is a 3x3 rotation matrix and ``t = [tx, ty, tz]`` is a 3x1
        translation vector. As rotation matrices must be proper orthogonal, the
        rotation component is orthonormalized using singular value decomposition
        before initialization.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> import numpy as np

        Creating a transform from a single matrix:

        >>> m = np.array([[0, 1, 0, 2],
        ...               [0, 0, 1, 3],
        ...               [1, 0, 0, 4],
        ...               [0, 0, 0, 1]])
        >>> tf = Tf.from_matrix(m)
        >>> tf.as_matrix()
        array([[0., 1., 0., 2.],
               [0., 0., 1., 3.],
               [1., 0., 0., 4.],
               [0., 0., 0., 1.]])
        >>> tf.single
        True

        Creating a transform from an N-dimensional array of matrices:

        >>> m = np.tile(np.eye(4), (2, 5, 1, 1))  # Shape (2, 5, 4, 4)
        >>> tf = Tf.from_matrix(m)
        >>> tf.shape
        (2, 5)
        >>> tf.single
        False
        >>> len(tf)
        2

        Matrices with a rotation component that is not proper orthogonal are
        orthogonalized using singular value decomposition before initialization:

        >>> tf = Tf.from_matrix(np.diag([2, 2, 2, 1]))
        >>> tf.as_matrix()
        array([[1., 0., 0., 0.],
               [0., 1., 0., 0.],
               [0., 0., 1., 0.],
               [0., 0., 0., 1.]])
        """
        return RigidTransform(matrix, normalize=True, copy=True)

    @staticmethod
    def from_rotation(rotation: Rotation) -> RigidTransform:
        """Initialize from a rotation, without a translation.

        When applying this transform to a vector ``v``, the result is the
        same as if the rotation was applied to the vector.
        ``Tf.from_rotation(r).apply(v) == r.apply(v)``

        Parameters
        ----------
        rotation : `Rotation` instance
            A single rotation or a rotation with N leading dimensions.

        Returns
        -------
        transform : `RigidTransform` instance
            Object containing the rigid transform represented by the input rotation.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> from scipy.spatial.transform import Rotation as R
        >>> import numpy as np

        Creating a transform from a single rotation:

        >>> r = R.from_euler("ZYX", [90, 30, 0], degrees=True)
        >>> r.apply([1, 0, 0])
        array([0.       , 0.8660254, -0.5     ])
        >>> tf = Tf.from_rotation(r)
        >>> tf.apply([1, 0, 0])
        array([0.       , 0.8660254, -0.5     ])
        >>> tf.single
        True

        The upper 3x3 submatrix of the transformation matrix is the rotation
        matrix:

        >>> np.allclose(tf.as_matrix()[:3, :3], r.as_matrix(), atol=1e-12)
        True

        Creating multiple transforms from a rotation with N leading dimensions:

        >>> r = R.from_euler("ZYX", [[90, 30, 0], [45, 30, 60]], degrees=True)
        >>> r.apply([1, 0, 0])
        array([[0.        , 0.8660254 , -0.5       ],
               [0.61237244, 0.61237244, -0.5       ]])
        >>> tf = Tf.from_rotation(r)
        >>> tf.apply([1, 0, 0])
        array([[0.        , 0.8660254 , -0.5       ],
               [0.61237244, 0.61237244, -0.5       ]])
        >>> tf.single
        False
        >>> len(tf)
        2
        """
        if not isinstance(rotation, Rotation):
            raise TypeError(
                "Expected `rotation` to be a `Rotation` instance, "
                f"got {type(rotation)}."
            )
        quat = rotation.as_quat()
        xp = array_namespace(quat)
        backend = select_backend(xp, quat.ndim < 3)
        matrix = backend.from_rotation(quat)
        return RigidTransform._from_raw_matrix(matrix, xp, backend)

    @staticmethod
    def from_translation(translation: ArrayLike) -> RigidTransform:
        """Initialize from a translation numpy array, without a rotation.

        When applying this transform to a vector ``v``, the result is the same
        as if the translation and vector were added together. If ``t`` is the
        displacement vector of the translation, then:

        ``Tf.from_translation(t).apply(v) == t + v``

        Parameters
        ----------
        translation : array_like, shape (..., 3)
            Translation vectors. Each translation[..., :] represents a 3D
            translation vector.

        Returns
        -------
        transform : `RigidTransform` instance
            Object containing the rigid transform represented by the input translation.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> import numpy as np

        Creating a transform from a single translation vector:

        >>> t = np.array([2, 3, 4])
        >>> t + np.array([1, 0, 0])
        array([3, 3, 4])
        >>> tf = Tf.from_translation(t)
        >>> tf.apply([1, 0, 0])
        array([3., 3., 4.])
        >>> tf.single
        True

        The top 3x1 points in the rightmost column of the transformation matrix
        is the translation vector:

        >>> tf.as_matrix()
        array([[1., 0., 0., 2.],
               [0., 1., 0., 3.],
               [0., 0., 1., 4.],
               [0., 0., 0., 1.]])
        >>> np.allclose(tf.as_matrix()[:3, 3], t)
        True

        Creating multiple transforms from an N-dimensional array of translation
        vectors:

        >>> t = np.array([[2, 3, 4], [1, 0, 0]])
        >>> t + np.array([1, 0, 0])
        array([[3, 3, 4],
               [2, 0, 0]])
        >>> tf = Tf.from_translation(t)
        >>> tf.apply([1, 0, 0])
        array([[3., 3., 4.],
               [2., 0., 0.]])
        >>> np.allclose(tf.as_matrix()[:, :3, 3], t)
        True
        >>> tf.single
        False
        >>> len(tf)
        2
        """
        xp = array_namespace(translation)
        translation = _promote(translation, xp=xp)
        backend = select_backend(xp, translation.ndim < 3)
        matrix = backend.from_translation(translation)
        return RigidTransform._from_raw_matrix(matrix, xp, backend)

    @staticmethod
    def from_components(translation: ArrayLike, rotation: Rotation) -> RigidTransform:
        """Initialize a rigid transform from translation and rotation
        components.

        When creating a rigid transform from a translation and rotation, the
        translation is applied after the rotation, such that
        ``tf = Tf.from_components(translation, rotation)``
        is equivalent to
        ``tf = Tf.from_translation(translation) * Tf.from_rotation(rotation)``.

        When applying a transform to a vector ``v``, the result is the
        same as if the transform was applied to the vector in the
        following way: ``tf.apply(v) == translation + rotation.apply(v)``

        Parameters
        ----------
        translation : array_like, shape (..., 3)
            Translation vectors. Each translation[..., :] represents a 3D
            translation vector.
        rotation : `Rotation` instance
            Rotation objects. The shape must be broadcastable with the
            translation shape.

        Returns
        -------
        transform : `RigidTransform` instance
            Rigid transform objects. The shape is determined by broadcasting
            the translation and rotation shapes together.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> from scipy.spatial.transform import Rotation as R
        >>> import numpy as np

        Creating from a single rotation and translation:

        >>> t = np.array([2, 3, 4])
        >>> r = R.from_euler("ZYX", [90, 30, 0], degrees=True)
        >>> r.as_matrix()
        array([[ 0.       , -1.,  0.        ],
               [ 0.8660254,  0.,  0.5       ],
               [-0.5      ,  0.,  0.8660254 ]])
        >>> tf = Tf.from_components(t, r)
        >>> tf.rotation.as_matrix()
        array([[ 0.       , -1.,  0.        ],
               [ 0.8660254,  0.,  0.5       ],
               [-0.5      ,  0.,  0.8660254 ]])
        >>> tf.translation
        array([2., 3., 4.])
        >>> tf.single
        True

        When applying a transform to a vector ``v``, the result is the same as
        if the transform was applied to the vector in the following way:
        ``tf.apply(v) == translation + rotation.apply(v)``

        >>> r.apply([1, 0, 0])
        array([0.       , 0.8660254, -0.5     ])
        >>> t + r.apply([1, 0, 0])
        array([2.       , 3.8660254,  3.5     ])
        >>> tf.apply([1, 0, 0])
        array([2.       , 3.8660254,  3.5     ])
        """
        rotation_tf = RigidTransform.from_rotation(rotation)
        return RigidTransform.from_translation(translation) * rotation_tf

    @staticmethod
    def from_exp_coords(exp_coords: ArrayLike) -> RigidTransform:
        r"""Initialize from exponential coordinates of transform.

        This implements the exponential map that converts 6-dimensional real
        vectors to SE(3).

        An exponential coordinate vector consists of 6 elements
        ``[rx, ry, rz, vx, vy, vz]``. The first 3 encode rotation (and form a
        rotation vector used in `Rotation.from_rotvec`) and the last 3 encode
        translation (and form a translation vector for pure translations).
        The exponential mapping can be expressed as matrix exponential
        ``T = exp(tau)``, where ``T`` is a 4x4 matrix representing a rigid
        transform and ``tau`` is a 4x4 matrix formed from the elements of an
        exponential coordinate vector::

            tau = [  0 -rz  ry vx]
                  [ rz   0 -rx vy]
                  [-ry  rx   0 vz]
                  [  0   0   0  1]

        Parameters
        ----------
        exp_coords : array_like, shape (..., 6)
            Exponential coordinate vectors. Each exp_coords[..., :] represents
            a 6D exponential coordinate vector with the expected order of
            components ``[rx, ry, rz, vx, vy, vz]``. The first 3 components
            encode rotation and the last 3 encode translation.

        Returns
        -------
        transform : `RigidTransform` instance
            Rigid transform objects with the same leading dimensions as the input.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> import numpy as np

        Creating from a single 6d vector of exponential coordinates:

        >>> tf = Tf.from_exp_coords([
        ...     -2.01041204, -0.52983629, 0.65773501,
        ...     0.10386614, 0.05855009, 0.54959179])
        >>> tf.as_matrix()
        array([[0.76406621, 0.10504613, -0.63652819, -0.10209961],
               [0.59956454, -0.47987325, 0.64050295, 0.40158789],
               [-0.2381705, -0.87102639, -0.42963687, 0.19637636],
               [0., 0., 0., 1.]])
        >>> tf.single
        True

        A vector of zeros represents the identity transform:

        >>> tf = Tf.from_exp_coords(np.zeros(6))
        >>> tf.as_matrix()
        array([[1., 0., 0., 0.],
               [0., 1., 0., 0.],
               [0., 0., 1., 0.],
               [0., 0., 0., 1.]])

        The last three numbers encode translation. If the first three numbers
        are zero, the last three components can be interpreted as the
        translation:

        >>> tf_trans = Tf.from_exp_coords([0, 0, 0, 4.3, -2, 3.4])
        >>> tf_trans.translation
        array([4.3, -2., 3.4])

        The first three numbers encode rotation as a rotation vector:

        >>> tf_rot = Tf.from_exp_coords([0.5, 0.3, 0.1, 0, 0, 0])
        >>> tf_rot.rotation.as_rotvec()
        array([0.5, 0.3, 0.1])

        Combining translation and rotation preserves the rotation vector,
        but changes the last three components as they encode translation and
        rotation:

        >>> (tf_trans * tf_rot).as_exp_coords()
        array([0.5, 0.3, 0.1, 3.64305882, -1.25879559, 4.46109265])
        """
        xp = array_namespace(exp_coords)
        exp_coords = xp_promote(exp_coords, force_floating=True, xp=xp)
        if exp_coords.shape[-1] != 6:
            raise ValueError(
                f"Expected `exp_coords` to have shape (..., 6), got {exp_coords.shape}."
            )
        backend = select_backend(xp, cython_compatible=exp_coords.ndim < 3)
        matrix = backend.from_exp_coords(exp_coords)
        return RigidTransform._from_raw_matrix(matrix, xp, backend)

    @staticmethod
    def from_dual_quat(
        dual_quat: ArrayLike, *, scalar_first: bool = False
    ) -> RigidTransform:
        """Initialize from a unit dual quaternion.

        Unit dual quaternions encode orientation in a real unit quaternion
        and translation in a dual quaternion. There is a double cover, i.e.,
        the unit dual quaternions q and -q represent the same transform.

        Unit dual quaternions must have a real quaternion with unit norm and
        a dual quaternion that is orthogonal to the real quaternion to satisfy
        the unit norm constraint. This function will enforce both properties
        through normalization.

        Parameters
        ----------
        dual_quat : array_like, shape (..., 8)
            Unit dual quaternions. Each dual_quat[..., :] represents a unit
            dual quaternion. The real part is stored in the first four
            components and the dual part in the last four components.
        scalar_first : bool, optional
            Whether the scalar component goes first or last in the two
            individual quaternions that represent the real and the dual part.
            Default is False, i.e. the scalar-last order is used.

        Returns
        -------
        transform : `RigidTransform` instance
            Rigid transform objects with the same leading dimensions as the input.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> import numpy as np

        Creating from a single unit dual quaternion:

        >>> tf = Tf.from_dual_quat([
        ...     0.0617101, -0.06483886, 0.31432811, 0.94508498,
        ...     0.04985168, -0.26119618, 0.1691491, -0.07743254])
        >>> tf.as_matrix()
        array([[0.79398752, -0.60213598, -0.08376202, 0.24605262],
               [0.58613113, 0.79477941, -0.15740392, -0.4932833],
               [0.16135089, 0.07588122, 0.98397557, 0.34262676],
               [0., 0., 0., 1.]])
        >>> tf.single
        True
        """
        xp = array_namespace(dual_quat)
        dual_quat = _promote(dual_quat, xp=xp)
        backend = select_backend(xp, dual_quat.ndim < 3)
        matrix = backend.from_dual_quat(dual_quat, scalar_first=scalar_first)
        return RigidTransform._from_raw_matrix(matrix, xp, backend)

    @staticmethod
    def identity(
        num: int | None = None, *, shape: int | tuple[int, ...] | None = None
    ) -> RigidTransform:
        """Initialize an identity transform.

        Composition with the identity transform has no effect, and
        applying the identity transform to a vector has no effect.

        Parameters
        ----------
        num : int, optional
            Number of identity transforms to generate. If None (default),
            then a single transform is generated.
        shape : int or tuple of ints, optional
            Shape of the identity transforms. If specified, `num` must
            be None.

        Returns
        -------
        transform : `RigidTransform` instance
            The identity transform.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> from scipy.spatial.transform import Rotation as R
        >>> import numpy as np

        Creating a single identity transform:

        >>> tf = Tf.identity()
        >>> tf.as_matrix()
        array([[1., 0., 0., 0.],
               [0., 1., 0., 0.],
               [0., 0., 1., 0.],
               [0., 0., 0., 1.]])
        >>> tf.single
        True

        The identity transform can be applied to a vector without effect:

        >>> tf.apply([1, 2, 3])
        array([1., 2., 3.])

        The identity transform when composed with another transform has no
        effect:

        >>> rng = np.random.default_rng(123)
        >>> t = rng.random(3)
        >>> r = R.random(rng=rng)
        >>> tf = Tf.from_components(t, r)
        >>> np.allclose((Tf.identity() * tf).as_matrix(),
        ...             tf.as_matrix(), atol=1e-12)
        True

        Multiple identity transforms can be generated at once:

        >>> tf = Tf.identity(2)
        >>> tf.as_matrix()
        array([[[1., 0., 0., 0.],
                [0., 1., 0., 0.],
                [0., 0., 1., 0.],
                [0., 0., 0., 1.]],
               [[1., 0., 0., 0.],
                [0., 1., 0., 0.],
                [0., 0., 1., 0.],
                [0., 0., 0., 1.]]])
        >>> tf.single
        False
        >>> len(tf)
        2
        """
        if num is not None and shape is not None:
            raise ValueError("Only one of `num` and `shape` can be specified.")
        if num is None and shape is None:
            shape = ()
        elif num is not None:
            shape = (num,)
        elif isinstance(shape, int):
            shape = (shape,)
        elif not isinstance(shape, tuple):
            raise ValueError("`shape` must be an int or a tuple of ints or None.")
        matrix = np.tile(np.eye(4), shape + (1, 1))
        # No need for a backend call here since identity is easy to construct and we are
        # currently not offering a backend-specific identity matrix
        return RigidTransform._from_raw_matrix(matrix, array_namespace(matrix))

    @staticmethod
    def concatenate(
        transforms: RigidTransform | Iterable[RigidTransform],
    ) -> RigidTransform:
        """
        Concatenate a sequence of `RigidTransform` objects into a
        single object.

        Parameters
        ----------
        transforms : sequence of `RigidTransform`
            If a single `RigidTransform` instance is passed in, a copy of
            it is returned.

        Returns
        -------
        transform : `RigidTransform` instance
            The concatenated transform.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> tf1 = Tf.from_translation([1, 0, 0])
        >>> tf2 = Tf.from_translation([[2, 0, 0], [3, 0, 0]])
        >>> Tf.concatenate([tf1, tf2]).translation
        array([[1., 0., 0.],
               [2., 0., 0.],
               [3., 0., 0.]])
        """
        if isinstance(transforms, RigidTransform):
            return RigidTransform._from_raw_matrix(
                transforms.as_matrix(), transforms._xp
            )
        if not all(isinstance(x, RigidTransform) for x in transforms):
            raise TypeError("input must contain RigidTransform objects only")

        xp = array_namespace(transforms[0].as_matrix())  # type:ignore[index]
        matrix = xp.concat(
            [xpx.atleast_nd(x.as_matrix(), ndim=3, xp=xp) for x in transforms]
        )
        return RigidTransform._from_raw_matrix(matrix, xp, None)

    def mean(self,
        weights: ArrayLike | None = None,
        axis: None | int | tuple[int, ...] = None
    ) -> RigidTransform:
        """Get the mean of the transforms.

        The mean of a set of transforms is the same as the mean of its
        rotation and translation components.

        The mean used for the rotation component is the chordal L2 mean (also
        called the projected or induced arithmetic mean) [1]_. If ``A`` is a
        set of rotation matrices, then the mean ``M`` is the rotation matrix
        that minimizes the following loss function:

        .. math::

            L(M) = \\sum_{i = 1}^{n} w_i \\lVert \\mathbf{A}_i -
            \\mathbf{M} \\rVert^2 ,

        where :math:`w_i`'s are the `weights` corresponding to each matrix.

        Parameters
        ----------
        weights : array_like shape (..., N), optional
            Weights describing the relative importance of the transforms. If
            None (default), then all values in `weights` are assumed to be
            equal. If given, the shape of `weights` must be broadcastable to
            the transform shape. Weights must be non-negative.
        axis : None, int, or tuple of ints, optional
            Axis or axes along which the means are computed. The default is to
            compute the mean of all transforms.

        Returns
        -------
        mean : `RigidTransform` instance
            Single transform containing the mean of the transforms in the
            current instance.

        References
        ----------
        .. [1] Hartley, Richard, et al.,
                "Rotation Averaging", International Journal of Computer Vision
                103, 2013, pp. 267-305.

        Examples
        --------
        >>> import numpy as np
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> from scipy.spatial.transform import Rotation as R
        >>> rng = np.random.default_rng(seed=123)

        The mean of a set of transforms is the same as the mean of the
        translation and rotation components:

        >>> t = rng.random((4, 3))
        >>> r = R.random(4, rng=rng)
        >>> tf = Tf.from_components(t, r)
        >>> tf.mean().as_matrix()
        array([[ 0.61593485, -0.74508342,  0.25588075,  0.66999034],
               [-0.59353615, -0.65246765, -0.47116962,  0.25481794],
               [ 0.51801458,  0.13833531, -0.84411151,  0.52429339],
               [0., 0., 0., 1.]])
        >>> Tf.from_components(t.mean(axis=0), r.mean()).as_matrix()
        array([[ 0.61593485, -0.74508342,  0.25588075,  0.66999034],
               [-0.59353615, -0.65246765, -0.47116962,  0.25481794],
               [ 0.51801458,  0.13833531, -0.84411151,  0.52429339],
               [0., 0., 0., 1.]])
        """
        mean = self._backend.mean(self._matrix, weights=weights, axis=axis)
        return RigidTransform._from_raw_matrix(mean, xp=self._xp,
                                               backend=self._backend)

    def as_matrix(self) -> Array:
        """Return a copy of the matrix representation of the transform.

        4x4 rigid transformation matrices are of the form::

            [       tx]
            [   R   ty]
            [       tz]
            [ 0 0 0  1]

        where ``R`` is a 3x3 orthonormal rotation matrix and
        ``t = [tx, ty, tz]`` is a 3x1 translation vector.

        Returns
        -------
        matrix : numpy.ndarray, shape (..., 4, 4)
            Transformation matrices with the same leading dimensions as the transform.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> from scipy.spatial.transform import Rotation as R
        >>> import numpy as np

        A transformation matrix is a 4x4 matrix formed from a 3x3 rotation
        matrix and a 3x1 translation vector:

        >>> t = np.array([2, 3, 4])
        >>> r = R.from_matrix([[0, 0, 1],
        ...                    [1, 0, 0],
        ...                    [0, 1, 0]])
        >>> tf = Tf.from_components(t, r)
        >>> tf.as_matrix()
        array([[ 0., 0., 1., 2.],
               [ 1., 0., 0., 3.],
               [ 0., 1., 0., 4.],
               [ 0., 0., 0., 1.]])

        >>> Tf.identity(2).as_matrix()
        array([[[1., 0., 0., 0.],
                [0., 1., 0., 0.],
                [0., 0., 1., 0.],
                [0., 0., 0., 1.]],
               [[1., 0., 0., 0.],
                [0., 1., 0., 0.],
                [0., 0., 1., 0.],
                [0., 0., 0., 1.]]])
        """
        matrix = self._xp.asarray(self._matrix, copy=True)
        if self._single:
            return matrix[0, ...]
        return matrix

    def as_components(self) -> tuple[Array, Rotation]:
        """Return the translation and rotation components of the transform,
        where the rotation is applied first, followed by the translation.

        4x4 rigid transformation matrices are of the form::

            [       tx]
            [   R   ty]
            [       tz]
            [ 0 0 0  1]

        Where ``R`` is a 3x3 orthonormal rotation matrix and
        ``t = [tx, ty, tz]`` is a 3x1 translation vector. This function
        returns the rotation corresponding to this rotation matrix
        ``r = Rotation.from_matrix(R)`` and the translation vector ``t``.

        When applying a transform ``tf`` to a vector ``v``, the result is the same
        as if the rotation and translation components were applied to the vector
        with the following operation:
        ``tf.apply(v) == translation + rotation.apply(v)``.

        Returns
        -------
        translation : numpy.ndarray, shape (..., 3)
            The translation of the transform.
        rotation : `Rotation` instance
            The rotation of the transform.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> from scipy.spatial.transform import Rotation as R
        >>> import numpy as np

        Recover the rotation and translation from a transform:

        >>> t = np.array([2, 3, 4])
        >>> r = R.from_matrix([[0, 0, 1],
        ...                    [1, 0, 0],
        ...                    [0, 1, 0]])
        >>> tf = Tf.from_components(t, r)
        >>> t, r = tf.as_components()
        >>> t
        array([2., 3., 4.])
        >>> r.as_matrix()
        array([[0., 0., 1.],
               [1., 0., 0.],
               [0., 1., 0.]])

        The transform applied to a vector is equivalent to the rotation applied
        to the vector, followed by the translation:

        >>> t + r.apply([1, 0, 0])
        array([2., 4., 4.])
        >>> tf.apply([1, 0, 0])
        array([2., 4., 4.])
        """
        return self.translation, self.rotation

    def as_exp_coords(self) -> Array:
        """Return the exponential coordinates of the transform.

        This implements the logarithmic map that converts SE(3) to 6-dimensional
        real vectors.

        This is an inverse of `from_exp_coords` where details on the mapping can
        be found.

        Returns
        -------
        exp_coords : numpy.ndarray, shape (..., 6)
            Exponential coordinate vectors with the same leading dimensions as the
            transform. The first three components define the rotation and the last
            three components define the translation.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> import numpy as np

        Get exponential coordinates of the identity matrix:

        >>> Tf.identity().as_exp_coords()
        array([0., 0., 0., 0., 0., 0.])
        """
        exp_coords = self._backend.as_exp_coords(self._matrix)
        if self._single:
            exp_coords = exp_coords[0, ...]
        return exp_coords

    def as_dual_quat(self, *, scalar_first: bool = False) -> Array:
        """Return the dual quaternion representation of the transform.

        Unit dual quaternions encode orientation in a real unit quaternion
        and translation in a dual quaternion. There is a double cover, i.e.,
        the unit dual quaternions q and -q represent the same transform.

        Parameters
        ----------
        scalar_first : bool, optional
            Whether the scalar component goes first or last in the two
            individual quaternions that represent the real and the dual part.
            Default is False, i.e. the scalar-last order is used.

        Returns
        -------
        dual_quat : numpy.ndarray, shape (..., 8)
            Unit dual quaternion vectors with the same leading dimensions as the
            transform. The real part is stored in the first four components and the
            dual part in the last four components.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> import numpy as np

        Get identity dual quaternion (we use scalar-last by default):

        >>> Tf.identity().as_dual_quat()
        array([0., 0., 0., 1., 0., 0., 0., 0.])

        When we want to use the scalar-first convention, we use the argument:

        >>> Tf.identity().as_dual_quat(scalar_first=True)
        array([1., 0., 0., 0., 0., 0., 0., 0.])
        """
        dual_quat = self._backend.as_dual_quat(self._matrix, scalar_first=scalar_first)
        if self._single:
            dual_quat = dual_quat[0]
        return dual_quat

    def __len__(self) -> int:
        """Return the length of the leading transform dimension.

        A transform can store an N-dimensional array of transforms. The length is the
        size of the first dimension of this array. If the transform is a single
        transform, the length is not defined and an error is raised.

        Returns
        -------
        length : int
            The number of transforms in this object.

        Raises
        ------
        TypeError
            If the transform is a single transform.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> import numpy as np
        >>> tf = Tf.identity(3)
        >>> len(tf)
        3

        An N-dimensional array of transforms returns its first dimension size:

        >>> t = np.ones((5, 2, 3, 1, 3))
        >>> tf = Tf.from_translation(t)
        >>> len(tf)
        5

        A single transform has no length:

        >>> tf = Tf.from_translation([1, 0, 0])
        >>> len(tf)  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        TypeError: Single transform has no len().
        """
        # We don't use self._single here because we also want to raise an error for
        # Array API backends that call len() on single RigidTransform objects.
        if self.single:
            raise TypeError("Single transform has no len")
        return self._matrix.shape[0]

    def __getitem__(
        self, indexer: int | slice | EllipsisType | None | ArrayLike
    ) -> RigidTransform:
        """Extract transform(s) at given index(es) from this object.

        Creates a new `RigidTransform` instance containing a subset of
        transforms stored in this object.

        Parameters
        ----------
        indexer : int or slice or array_like
            Specifies which transform(s) to extract. A single indexer must be
            specified, i.e. as if indexing a 1 dimensional array or list.

        Returns
        -------
        transform : `RigidTransform` instance
            Contains
                - a single transform, if `indexer` is a single index
                - a stack of transform(s), if `indexer` is a slice, or an index
                  array.

        Raises
        ------
        TypeError
            If the transform is a single transform.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> t = [[0, 0, 0], [1, 0, 0], [2, 0, 0]]  # 3 translations
        >>> tf = Tf.from_translation(t)

        A single index returns a single transform:

        >>> tf[0].as_matrix()
        array([[1., 0., 0., 0.],
               [0., 1., 0., 0.],
               [0., 0., 1., 0.],
               [0., 0., 0., 1.]])

        A slice returns a stack of transforms:

        >>> tf[1:3].translation
        array([[1., 0., 0.],
               [2., 0., 0.]])

        An index array returns a stack of transforms:

        >>> tf[[0, 2]].translation
        array([[0., 0., 0.],
               [2., 0., 0.]])
        """
        if self.single:
            raise TypeError("Single transform is not subscriptable.")

        is_array = isinstance(indexer, type(self._matrix))
        xp = self._xp
        # Masking is only specified in the Array API when the array is the sole index
        # This special case handling is necessary to support boolean indexing and
        # integer array indexing with take (see
        # https://github.com/data-apis/array-api/pull/900#issuecomment-2674432480)
        # Ideally we would converge to [indexer, ...] indexing, but this is not
        # supported for now.
        if is_array and indexer.dtype == xp.bool:  # type:ignore[union-attr]
            return RigidTransform(self._matrix[indexer], normalize=False)
        if is_array and xp.isdtype(indexer.dtype, "integral"):  # type:ignore[union-attr]
            if self._matrix.shape[0] == 0:
                raise IndexError("cannot take from an empty array")
            return RigidTransform(
                xp.take(self._matrix, indexer, axis=0), normalize=False
            )
        return RigidTransform(self._matrix[indexer, ...], normalize=False)

    def __setitem__(
        self,
        indexer: int | slice | EllipsisType | None | ArrayLike,
        value: RigidTransform,
    ):
        """Set transform(s) at given index(es) in this object.

        Parameters
        ----------
        indexer : int or slice or array_like
            Specifies which transform(s) to replace. A single indexer must be
            specified, i.e. as if indexing a 1 dimensional array or list.

        value : `RigidTransform` instance
            The transform(s) to set.

        Raises
        ------
        TypeError
            If the transform is a single transform.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> t = [[0, 0, 0], [1, 0, 0], [2, 0, 0]]  # 3 translations
        >>> tf = Tf.from_translation(t)

        Set a single transform:

        >>> tf[0] = Tf.from_translation([9, 9, 9])
        >>> tf.translation
        array([[9., 9., 9.],
               [1., 0., 0.],
               [2., 0., 0.]])
        """
        if self.single:
            raise TypeError("Single transform is not subscriptable.")

        if not isinstance(value, RigidTransform):
            raise TypeError("value must be a RigidTransform object")

        self._matrix = self._backend.setitem(self._matrix, indexer, value.as_matrix())

    def __mul__(
        self, other: RigidTransform | Rotation
    ) -> RigidTransform | NotImplementedType:
        """Compose this transform with the other.

        If ``p`` and ``q`` are two transforms, then the composition of '``q``
        followed by ``p``' is equivalent to ``p * q``. In terms of
        transformation matrices, the composition can be expressed as
        ``p.as_matrix() @ q.as_matrix()``.

        In terms of translations and rotations, the composition when applied to
        a vector ``v`` is equivalent to
        ``p.translation + p.rotation.apply(q.translation)
        + (p.rotation * q.rotation).apply(v)``.

        This function supports composition of multiple transforms at a time using
        broadcasting rules. The resulting shape for two `RigidTransform` instances
        ``p`` and ``q`` is `np.broadcast_shapes(p.shape, q.shape)`.

        If ``other`` is a `Rotation`, it is automatically promoted to a
        `RigidTransform` with zero translation.

        Parameters
        ----------
        other : `RigidTransform` or `Rotation` instance
            Transform(s) or rotation(s) to be composed with this one. The shapes
            must be broadcastable.

        Returns
        -------
        `RigidTransform` instance
            The composed transform.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> from scipy.spatial.transform import Rotation as R
        >>> import numpy as np

        Compose two transforms:

        >>> tf1 = Tf.from_translation([1, 0, 0])
        >>> tf2 = Tf.from_translation([0, 1, 0])
        >>> tf = tf1 * tf2
        >>> tf.translation
        array([1., 1., 0.])
        >>> tf.single
        True

        When applied to a vector, the composition of two transforms is applied
        in right-to-left order.

        >>> t1, r1 = [1, 2, 3], R.from_euler('z', 60, degrees=True)
        >>> t2, r2 = [0, 1, 0], R.from_euler('x', 30, degrees=True)
        >>> tf1 = Tf.from_components(t1, r1)
        >>> tf2 = Tf.from_components(t2, r2)
        >>> tf = tf1 * tf2
        >>> tf.apply([1, 0, 0])
        array([0.6339746, 3.3660254, 3.       ])
        >>> tf1.apply(tf2.apply([1, 0, 0]))
        array([0.6339746, 3.3660254, 3.       ])

        When at least one of the transforms is not single, the result is a stack
        of transforms.

        >>> tf1 = Tf.from_translation([1, 0, 0])
        >>> tf2 = Tf.from_translation([[0, 2, 0], [0, 0, 3]])
        >>> tf = tf1 * tf2
        >>> tf.translation
        array([[1., 2., 0.],
               [1., 0., 3.]])
        >>> tf.single
        False
        >>> len(tf)
        2

        Broadcasting rules apply when composing multiple transforms at a time.

        >>> tf1 = Tf.from_translation(np.ones((5, 1, 3)))  # Shape (5, 1, 3)
        >>> tf2 = Tf.from_translation(np.ones((1, 4, 3)))  # Shape (1, 4, 3)
        >>> tf = tf1 * tf2  # Shape (5, 4, 3)
        >>> tf.translation.shape
        (5, 4, 3)
        """
        if isinstance(other, Rotation):
            other = RigidTransform.from_rotation(other)
        elif not isinstance(other, RigidTransform):
            # If other is not a RigidTransform, we return NotImplemented to allow other
            # types to implement __rmul__
            return NotImplemented
        if not broadcastable(self._matrix.shape, other._matrix.shape):
            raise ValueError(
                f"Cannot broadcast {self._matrix.shape[:-2]} transforms in "
                f"first to {other._matrix.shape[:-2]} transforms in second object."
            )
        cython_compatible = self._matrix.ndim < 4 and other._matrix.ndim < 4
        backend = select_backend(self._xp, cython_compatible=cython_compatible)
        matrix = backend.compose_transforms(self._matrix, other._matrix)
        # Only necessary for cython. Array API broadcasting handles this by default
        if self._single and other._single:
            matrix = matrix[0, ...]
        return RigidTransform(matrix, normalize=True, copy=False)

    def __rmul__(self, other: Rotation) -> RigidTransform | NotImplementedType:
        """Compose a rotation with this transform (rotation applied second).

        See `__mul__` for more details.

        Parameters
        ----------
        other : `Rotation` instance
            The rotation to compose with this transform. The shapes must be
            broadcastable.

        Returns
        -------
        `RigidTransform` instance
            The composed transform.
        """
        if isinstance(other, Rotation):
            return RigidTransform.from_rotation(other) * self
        # When other is a RigidTransform __mul__ is called, so we don't handle it here
        return NotImplemented

    def __pow__(self, n: float) -> RigidTransform:
        """Compose this transform with itself `n` times.

        A rigid transform `p` when raised to non-integer powers can be thought
        of as finding a fraction of the transformation. For example, a power of
        0.5 finds a "halfway" transform from the identity to `p`.

        This is implemented by applying screw linear interpolation (ScLERP)
        between `p` and the identity transform, where the angle of the rotation
        component is scaled by `n`, and the translation is proportionally
        adjusted along the screw axis.

        ``q = p ** n`` can also be expressed as
        ``q = RigidTransform.from_exp_coords(p.as_exp_coords() * n)``.

        If `n` is negative, then the transform is inverted before the power
        is applied. In other words, ``p ** -abs(n) == p.inv() ** abs(n)``.

        Parameters
        ----------
        n : float
            The number of times to compose the transform with itself.

        Returns
        -------
        `RigidTransform` instance
            If the input Rotation `p` contains `N` multiple rotations, then
            the output will contain `N` rotations where the `i` th rotation
            is equal to ``p[i] ** n``.

        Notes
        -----
        There are three notable cases: if ``n == 1`` then a copy of the original
        transform is returned, if ``n == 0`` then the identity transform is
        returned, and if ``n == -1`` then the inverse transform is returned.

        Note that fractional powers ``n`` which effectively take a root of
        rotation, do so using the shortest path smallest representation of that
        angle (the principal root). This means that powers of ``n`` and ``1/n``
        are not necessarily inverses of each other. For example, a 0.5 power of
        a +240 degree rotation will be calculated as the 0.5 power of a -120
        degree rotation, with the result being a rotation of -60 rather than
        +120 degrees.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> import numpy as np

        A power of 2 returns the transform composed with itself:

        >>> tf = Tf.from_translation([1, 2, 3])
        >>> (tf ** 2).translation
        array([2., 4., 6.])
        >>> (tf ** 2).as_matrix()
        array([[1., 0., 0., 2.],
               [0., 1., 0., 4.],
               [0., 0., 1., 6.],
               [0., 0., 0., 1.]])

        A negative power returns the inverse of the transform raised to the
        absolute value of `n`:

        >>> (tf ** -2).translation
        array([-2., -4., -6.])
        >>> np.allclose((tf ** -2).as_matrix(), (tf.inv() ** 2).as_matrix(),
        ...             atol=1e-12)
        True

        A power of 0 returns the identity transform:

        >>> (tf ** 0).as_matrix()
        array([[1., 0., 0., 0.],
               [0., 1., 0., 0.],
               [0., 0., 1., 0.],
               [0., 0., 0., 1.]])

        A power of 1 returns a copy of the original transform:

        >>> (tf ** 1).as_matrix()
        array([[1., 0., 0., 1.],
               [0., 1., 0., 2.],
               [0., 0., 1., 3.],
               [0., 0., 0., 1.]])

        A fractional power returns a transform with a scaled rotation and
        translated along the screw axis. Here we take the square root of the
        transform, which when squared recovers the original transform:

        >>> tf_half = (tf ** 0.5)
        >>> tf_half.translation
        array([0.5, 1., 1.5])
        >>> (tf_half ** 2).as_matrix()
        array([[1., 0., 0., 1.],
               [0., 1., 0., 2.],
               [0., 0., 1., 3.],
               [0., 0., 0., 1.]])
        """
        matrix = self._backend.pow(self._matrix, n)
        if self._single:
            matrix = matrix[0, ...]
        return RigidTransform._from_raw_matrix(matrix, self._xp, self._backend)

    def inv(self) -> RigidTransform:
        """Invert this transform.

        Composition of a transform with its inverse results in an identity
        transform.

        A rigid transform is a composition of a rotation and a translation,
        where the rotation is applied first, followed by the translation. So the
        inverse transform is equivalent to the inverse translation followed by
        the inverse rotation.

        Returns
        -------
        `RigidTransform` instance
            The inverse of this transform.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> from scipy.spatial.transform import Rotation as R
        >>> import numpy as np

        A transform composed with its inverse results in an identity transform:

        >>> rng = np.random.default_rng(seed=123)
        >>> t = rng.random(3)
        >>> r = R.random(rng=rng)
        >>> tf = Tf.from_components(t, r)
        >>> tf.as_matrix()
        array([[-0.45431291,  0.67276178, -0.58394466,  0.68235186],
               [-0.23272031,  0.54310598,  0.80676958,  0.05382102],
               [ 0.85990758,  0.50242162, -0.09017473,  0.22035987],
               [ 0.        ,  0.        ,  0.        ,  1.        ]])

        >>> (tf.inv() * tf).as_matrix()
        array([[[1., 0., 0., 0.],
                [0., 1., 0., 0.],
                [0., 0., 1., 0.],
                [0., 0., 0., 1.]]])

        The inverse rigid transform is the same as the inverse translation
        followed by the inverse rotation:

        >>> t, r = tf.as_components()
        >>> r_inv = r.inv()  # inverse rotation
        >>> t_inv = -t  # inverse translation
        >>> tf_r_inv = Tf.from_rotation(r_inv)
        >>> tf_t_inv = Tf.from_translation(t_inv)
        >>> np.allclose((tf_r_inv * tf_t_inv).as_matrix(),
        ...             tf.inv().as_matrix(),
        ...             atol=1e-12)
        True
        >>> (tf_r_inv * tf_t_inv * tf).as_matrix()
        array([[[1., 0., 0., 0.],
                [0., 1., 0., 0.],
                [0., 0., 1., 0.],
                [0., 0., 0., 1.]]])
        """
        matrix = self._backend.inv(self._matrix)
        if self._single:
            matrix = matrix[0, ...]
        return RigidTransform._from_raw_matrix(matrix, self._xp, self._backend)

    def apply(self, vector: ArrayLike, inverse: bool = False) -> Array:
        """Apply the transform to a vector.

        If the original frame transforms to the final frame by this transform,
        then its application to a vector can be seen in two ways:

            - As a projection of vector components expressed in the final frame
              to the original frame.
            - As the physical transformation of a vector being glued to the
              original frame as it transforms. In this case the vector
              components are expressed in the original frame before and after
              the transformation.

        In terms of the rotation matrix and translation,
        this application is the same as
        ``self.translation + (mat @ vector[..., np.newaxis])[..., 0]``,
        where ``mat = self.rotation.as_matrix()``.

        For a single transform, this is the same as
        ``self.translation + vector @ self.rotation.as_matrix().T``.

        Parameters
        ----------
        vector : array_like, shape (..., 3)
            Vector(s) to be transformed. Each vector[..., :] represents a 3D
            vector.
        inverse : bool, optional
            If True, the inverse of the transform is applied to the vector.

        Returns
        -------
        transformed_vector : numpy.ndarray, shape (..., 3)
            The transformed vector(s) with shape determined by broadcasting
            the transform and vector shapes together.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> from scipy.spatial.transform import Rotation as R
        >>> import numpy as np

        Apply a single transform to a vector. Here the transform is just a
        translation, so the result is the vector added to the translation
        vector.

        >>> t = np.array([1, 2, 3])
        >>> tf = Tf.from_translation(t)
        >>> t + np.array([1, 0, 0])
        array([2, 2, 3])
        >>> tf.apply([1, 0, 0])
        array([2., 2., 3.])

        Apply a single transform to a stack of vectors:

        >>> tf.apply([[1, 0, 0], [0, 1, 0]])
        array([[2., 2., 3.],
               [1., 3., 3.]])

        Apply the inverse of a transform to a vector, so the result is the
        negative of the translation vector added to the vector.

        >>> -t + np.array([1, 0, 0])
        array([0, -2, -3])
        >>> tf.apply([1, 0, 0], inverse=True)
        array([0., -2., -3.])

        Broadcasting is supported when applying multiple transforms to an N-dimensional
        array of vectors.

        >>> tf = Tf.from_translation(np.ones((4, 5, 1, 3)))
        >>> vectors = np.zeros((2, 3))
        >>> tf.apply(vectors).shape
        (4, 5, 2, 3)

        For transforms which are not just pure translations, applying it to a
        vector is the same as applying the rotation component to the vector and
        then adding the translation component.

        >>> r = R.from_euler('z', 60, degrees=True)
        >>> tf = Tf.from_components(t, r)
        >>> t + r.apply([1, 0, 0])
        array([1.5,       2.8660254, 3.       ])
        >>> tf.apply([1, 0, 0])
        array([1.5,       2.8660254, 3.       ])

        When applying the inverse of a transform, the result is the negative of
        the translation vector added to the vector, and then rotated by the
        inverse rotation.

        >>> r.inv().apply(-t + np.array([1, 0, 0]))
        array([-1.73205081, -1.        , -3.        ])
        >>> tf.apply([1, 0, 0], inverse=True)
        array([-1.73205081, -1.        , -3.        ])
        """
        # We do not use the cached xp here to catch cases where matrix and vector have
        # different xp instances.
        xp = array_namespace(self._matrix, vector)
        vector = xp.asarray(
            vector, dtype=self._matrix.dtype, device=device(self._matrix)
        )
        cython_compatible = self._matrix.ndim < 4 and vector.ndim < 3
        backend = select_backend(xp, cython_compatible=cython_compatible)
        result = backend.apply(self._matrix, vector, inverse)
        if self._single and vector.ndim == 1:
            result = result[0, ...]
        return result

    @property
    def rotation(self) -> Rotation:
        """Return the rotation component of the transform.

        A transform is a composition of a rotation and a translation, such that
        when applied to a vector, the vector is first rotated and then
        translated. This property returns the rotation part of the transform.

        Returns
        -------
        rotation : `Rotation` instance
            A single rotation or a stack of rotations.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> from scipy.spatial.transform import Rotation as R
        >>> import numpy as np

        The rotation component is extracted from the transform:

        >>> t = np.array([1, 0, 0])
        >>> r = R.random(3)
        >>> tf = Tf.from_components(t, r)
        >>> np.allclose(tf.rotation.as_matrix(), r.as_matrix())
        True
        """
        if self._single:
            return Rotation.from_matrix(self._matrix[0, :3, :3])
        return Rotation.from_matrix(self._matrix[..., :3, :3])

    @property
    def translation(self) -> Array:
        """Return the translation component of the transform.

        A transform is a composition of a rotation and a translation, such that
        when applied to a vector, the vector is first rotated and then
        translated. This property returns the translation part of the transform.

        Returns
        -------
        translation : numpy.ndarray, shape (..., 3)
            Translation vectors with the same leading dimensions as the transform.

        Examples
        --------
        >>> from scipy.spatial.transform import RigidTransform as Tf
        >>> from scipy.spatial.transform import Rotation as R
        >>> import numpy as np

        The translation component is extracted from the transform:

        >>> t = np.array([[1, 0, 0], [2, 0, 0], [3, 0, 0]])
        >>> r = R.random()
        >>> tf = Tf.from_components(t, r)
        >>> np.allclose(tf.translation, t)
        True
        """
        if self._single:
            return self._xp.asarray(self._matrix[0, :3, 3], copy=True)
        return self._xp.asarray(self._matrix[..., :3, 3], copy=True)

    @property
    def single(self) -> bool:
        """Whether this instance represents a single transform.

        Single transforms are not subscriptable, and do not have a length.

        Returns
        -------
        single : bool
            True if this instance represents a single transform, False
            otherwise.
        """
        return self._single or self._matrix.ndim == 2

    @property
    def shape(self) -> tuple[int, ...]:
        """The shape of the transform's leading dimensions."""
        if self._single or self._matrix.ndim == 2:
            return ()
        return self._matrix.shape[:-2]

    def __reduce__(self) -> tuple[Callable, tuple]:
        """Reduce the RigidTransform for pickling.

        We store modules inside RigidTransforms which cannot be pickled. To circumvent
        this, we pickle only the matrix and restore the cached modules from the matrix
        type in `from_matrix`.
        """
        matrix = self._matrix
        if self._single:
            matrix = matrix[0, ...]
        return (self.__class__.from_matrix, (matrix,))

    def __iter__(self) -> Iterator[RigidTransform]:
        """Iterate over transforms."""
        if self._single or self._matrix.ndim == 2:
            raise TypeError("Single transform is not iterable.")
        # We return a generator that yields a new RigidTransform object for each
        # transform in the current object. We cannot rely on the default implementation
        # because jax will not raise an IndexError for out-of-bounds indices.
        for i in range(self._matrix.shape[0]):
            yield RigidTransform._from_raw_matrix(
                self._matrix[i, ...], self._xp, self._backend
            )

    @staticmethod
    def _from_raw_matrix(
        matrix: Array, xp: ModuleType, backend: ModuleType | None = None
    ) -> RigidTransform:
        """Create a RigidTransform skipping all sanitization steps.

        This method is is intended for internal, performant creation of RigidTransforms
        with matrices that are guaranteed to be valid.
        """
        tf = RigidTransform.__new__(RigidTransform)
        tf._single = matrix.ndim == 2 and is_numpy(xp)
        if tf._single:
            matrix = xpx.atleast_nd(matrix, ndim=3, xp=xp)
        tf._matrix = matrix
        tf._xp = xp
        if backend is None:
            backend = select_backend(xp, matrix.ndim < 4)
        tf._backend = backend
        return tf



"""
Unit tests for trust-region optimization routines.

"""
import pytest
import numpy as np
from numpy.testing import assert_, assert_equal, assert_allclose
from scipy.optimize import (minimize, rosen, rosen_der, rosen_hess,
                            rosen_hess_prod)


class Accumulator:
    """ This is for testing callbacks."""
    def __init__(self):
        self.count = 0
        self.accum = None

    def __call__(self, x):
        self.count += 1
        if self.accum is None:
            self.accum = np.array(x)
        else:
            self.accum += x


class TestTrustRegionSolvers:

    def setup_method(self):
        self.x_opt = [1.0, 1.0]
        self.easy_guess = [2.0, 2.0]
        self.hard_guess = [-1.2, 1.0]

    def test_dogleg_accuracy(self):
        # test the accuracy and the return_all option
        x0 = self.hard_guess
        r = minimize(rosen, x0, jac=rosen_der, hess=rosen_hess, tol=1e-8,
                     method='dogleg', options={'return_all': True},)
        assert_allclose(x0, r['allvecs'][0])
        assert_allclose(r['x'], r['allvecs'][-1])
        assert_allclose(r['x'], self.x_opt)

    def test_dogleg_callback(self):
        # test the callback mechanism and the maxiter and return_all options
        accumulator = Accumulator()
        maxiter = 5
        r = minimize(rosen, self.hard_guess, jac=rosen_der, hess=rosen_hess,
                     callback=accumulator, method='dogleg',
                     options={'return_all': True, 'maxiter': maxiter},)
        assert_equal(accumulator.count, maxiter)
        assert_equal(len(r['allvecs']), maxiter+1)
        assert_allclose(r['x'], r['allvecs'][-1])
        assert_allclose(sum(r['allvecs'][1:]), accumulator.accum)

    def test_dogleg_user_warning(self):
        with pytest.warns(RuntimeWarning,
                          match=r'Maximum number of iterations'):
            minimize(rosen, self.hard_guess, jac=rosen_der,
                     hess=rosen_hess, method='dogleg',
                     options={'disp': True, 'maxiter': 1}, )

    def test_solver_concordance(self):
        # Assert that dogleg uses fewer iterations than ncg on the Rosenbrock
        # test function, although this does not necessarily mean
        # that dogleg is faster or better than ncg even for this function
        # and especially not for other test functions.
        f = rosen
        g = rosen_der
        h = rosen_hess
        for x0 in (self.easy_guess, self.hard_guess):
            r_dogleg = minimize(f, x0, jac=g, hess=h, tol=1e-8,
                                method='dogleg', options={'return_all': True})
            r_trust_ncg = minimize(f, x0, jac=g, hess=h, tol=1e-8,
                                   method='trust-ncg',
                                   options={'return_all': True})
            r_trust_krylov = minimize(f, x0, jac=g, hess=h, tol=1e-8,
                                   method='trust-krylov',
                                   options={'return_all': True})
            r_ncg = minimize(f, x0, jac=g, hess=h, tol=1e-8,
                             method='newton-cg', options={'return_all': True})
            r_iterative = minimize(f, x0, jac=g, hess=h, tol=1e-8,
                                   method='trust-exact',
                                   options={'return_all': True})
            assert_allclose(self.x_opt, r_dogleg['x'])
            assert_allclose(self.x_opt, r_trust_ncg['x'])
            assert_allclose(self.x_opt, r_trust_krylov['x'])
            assert_allclose(self.x_opt, r_ncg['x'])
            assert_allclose(self.x_opt, r_iterative['x'])
            assert_(len(r_dogleg['allvecs']) < len(r_ncg['allvecs']))

    def test_trust_ncg_hessp(self):
        for x0 in (self.easy_guess, self.hard_guess, self.x_opt):
            r = minimize(rosen, x0, jac=rosen_der, hessp=rosen_hess_prod,
                         tol=1e-8, method='trust-ncg')
            assert_allclose(self.x_opt, r['x'])

    def test_trust_ncg_start_in_optimum(self):
        r = minimize(rosen, x0=self.x_opt, jac=rosen_der, hess=rosen_hess,
                     tol=1e-8, method='trust-ncg')
        assert_allclose(self.x_opt, r['x'])

    def test_trust_krylov_start_in_optimum(self):
        r = minimize(rosen, x0=self.x_opt, jac=rosen_der, hess=rosen_hess,
                     tol=1e-8, method='trust-krylov')
        assert_allclose(self.x_opt, r['x'])

    def test_trust_exact_start_in_optimum(self):
        r = minimize(rosen, x0=self.x_opt, jac=rosen_der, hess=rosen_hess,
                     tol=1e-8, method='trust-exact')
        assert_allclose(self.x_opt, r['x'])



"""PUBLIC_MODULES was once included in scipy._lib.tests.test_public_api.

It has been separated into this file so that this list of public modules
could be used when generating tables showing support for alternative
array API backends across modules in
scipy/doc/source/array_api_capabilities.py.
"""

# Historically SciPy has not used leading underscores for private submodules
# much.  This has resulted in lots of things that look like public modules
# (i.e. things that can be imported as `import scipy.somesubmodule.somefile`),
# but were never intended to be public.  The PUBLIC_MODULES list contains
import pathlib

# modules that are either public because they were meant to be, or because they
# contain public functions/objects that aren't present in any other namespace
# for whatever reason and therefore should be treated as public.
PUBLIC_MODULES = ["scipy." + s for s in [
    "cluster",
    "cluster.vq",
    "cluster.hierarchy",
    "constants",
    "datasets",
    "differentiate",
    "fft",
    "fftpack",
    "integrate",
    "interpolate",
    "io",
    "io.arff",
    "io.matlab",
    "io.wavfile",
    "linalg",
    "linalg.blas",
    "linalg.cython_blas",
    "linalg.lapack",
    "linalg.cython_lapack",
    "linalg.interpolative",
    "ndimage",
    "odr",
    "optimize",
    "optimize.elementwise",
    "signal",
    "signal.windows",
    "sparse",
    "sparse.linalg",
    "sparse.csgraph",
    "spatial",
    "spatial.distance",
    "spatial.transform",
    "special",
    "stats",
    "stats.contingency",
    "stats.distributions",
    "stats.mstats",
    "stats.qmc",
    "stats.sampling"
]]

# Handle the `_without-fortran` build option
_without_fortran = False
if not (pathlib.Path(__file__).parent.parent / 'odr').is_dir():
    _without_fortran = True
    PUBLIC_MODULES.remove('scipy.odr')



""" Test functions for the sparse.linalg._isolve module
"""

from contextlib import nullcontext
import itertools
import platform
import pytest
import types

import numpy as np
from numpy.testing import assert_allclose
from numpy import zeros, arange, array, ones, eye, iscomplexobj

from scipy._lib._array_api import (
    is_numpy, make_xp_pytest_param, make_xp_pytest_marks,
    xp_assert_close, xp_assert_less, xp_vector_norm, xp_assert_equal,
    xp_assert_less_equal,
)
from scipy._external import array_api_extra as xpx
from scipy._lib._sparse import issparse
from scipy.sparse import dia_array, csr_array, kronsum

from scipy.sparse.linalg import LinearOperator, aslinearoperator
from scipy.sparse.linalg._isolve import (bicg, bicgstab, cg, cgs,
                                         gcrotmk, gmres, lgmres,
                                         minres, qmr, tfqmr)
from scipy.sparse.linalg._interface import IdentityOperator

# TODO check that method preserve shape and type
# TODO test both preconditioner methods


# list of all solvers under test
_SOLVERS = [bicg, bicgstab, cg, cgs, gcrotmk, gmres, lgmres,
            minres, qmr, tfqmr]

CB_TYPE_FILTER = ".*called without specifying `callback_type`.*"


def _assert_success(*, A, x, b, xp, rtol=1.0, atol=0.0, less_equal=False):
    Ax = xp.squeeze(A @ x[..., xp.newaxis], axis=-1)
    residual = Ax - b
    err = xp_vector_norm(residual, axis=-1)
    assertion = xp_assert_less_equal if less_equal else xp_assert_less
    limit = atol + rtol * xp_vector_norm(b, axis=-1)
    assertion(err, limit)


# create parametrized fixture for easy reuse in tests
@pytest.fixture(
    scope="session",
    params=[make_xp_pytest_param(func) for func in _SOLVERS],
)
def solver(request):
    """
    Fixture for all solvers in scipy.sparse.linalg._isolve
    """
    return request.param


class Case:
    def __init__(self, name, A, b=None, skip=None, nonconvergence=None):
        self.name = name
        self.A = A
        if b is None:
            self.b = arange(A.shape[0], dtype=A.dtype)
        else:
            self.b = b
        if skip is None:
            self.skip = []
        else:
            self.skip = skip
        if nonconvergence is None:
            self.nonconvergence = []
        else:
            self.nonconvergence = nonconvergence


class SingleTest:
    # case with a specific solver
    def __init__(self, A, b, solver, casename, convergence=True):
        self.A = A
        self.b = b
        self.solver = solver
        self.casename = casename
        self.name = casename + '-' + solver.__name__
        self.convergence = convergence

    def __repr__(self):
        return f"<{self.name}>"


def xp_case(case, xp, batch_A, batch_b, rng=None):
    sparse = issparse(case.A)

    if (not_np := not is_numpy(xp)) and sparse: # TODO: pydata/sparse for sparse tests?
        pytest.skip("sparse tests run only with `np` backend")

    A = xp.asarray(case.A) if not_np else case.A
    b = xp.asarray(case.b) if not_np else case.b

    rng = np.random.default_rng(rng)
    if sparse and (batch_A or batch_b):
        A = A.tocoo()
    if batch_A:
        # apply some random scaling to each system in batch,
        # preserving symmetry, (non-)positive definiteness, Hermiticity
        scaling = np.abs(rng.standard_normal(batch_A + (1, 1)))
        batch_array = xp.asarray(scaling, dtype=A.dtype)
        A = batch_array * A
    if batch_b:
        # apply some random scaling to each RHS in batch
        b = b * xp.asarray(rng.random(batch_b + (1,)), dtype=b.dtype)

    return types.SimpleNamespace(
        name=case.name,
        A=A,
        b=b,
        convergence=case.convergence,
        solver=case.solver,
        casename=case.casename,
    )


class IterativeParams:
    def __init__(self):
        sym_solvers = [minres, cg]
        posdef_solvers = [cg]
        real_solvers = [minres]

        # list of Cases
        self.cases = []

        # Symmetric and Positive Definite
        N = 40
        data = ones((3, N))
        data[0, :] = 2
        data[1, :] = -1
        data[2, :] = -1
        Poisson1D = dia_array((data, [0, -1, 1]), shape=(N, N)).tocsr()
        self.cases.append(Case("poisson1d", Poisson1D))
        # note: minres fails for single precision
        self.cases.append(Case("poisson1d-F", Poisson1D.astype(np.float32),
                               skip=[minres]))

        # Symmetric and Negative Definite
        self.cases.append(Case("neg-poisson1d", -Poisson1D,
                               skip=posdef_solvers))
        # note: minres fails for single precision
        self.cases.append(Case("neg-poisson1d-F", (-Poisson1D).astype('f'),
                               skip=posdef_solvers + [minres]))

        # 2-dimensional Poisson equations
        Poisson2D = kronsum(Poisson1D, Poisson1D)
        # note: minres fails for 2-d poisson problem,
        # it will be fixed in the future PR
        self.cases.append(Case("poisson2d", Poisson2D, skip=[minres]))
        # note: minres fails for single precision
        self.cases.append(Case("poisson2d-F", Poisson2D.astype('f'),
                               skip=[minres]))

        # Symmetric and Indefinite
        data = array([[6, -5, 2, 7, -1, 10, 4, -3, -8, 9]], dtype='d')
        RandDiag = dia_array((data, [0]), shape=(10, 10)).tocsr()
        self.cases.append(Case("rand-diag", RandDiag, skip=posdef_solvers))
        self.cases.append(Case("rand-diag-F", RandDiag.astype('f'),
                               skip=posdef_solvers))

        # Random real-valued
        rng = np.random.RandomState(1234)
        data = rng.rand(4, 4)
        self.cases.append(Case("rand", data,
                               skip=posdef_solvers + sym_solvers))
        self.cases.append(Case("rand-F", data.astype('f'),
                               skip=posdef_solvers + sym_solvers))

        # Random symmetric real-valued
        rng = np.random.RandomState(1234)
        data = rng.rand(4, 4)
        data = data + data.T
        self.cases.append(Case("rand-sym", data, skip=posdef_solvers))
        self.cases.append(Case("rand-sym-F", data.astype('f'),
                               skip=posdef_solvers))

        # Random pos-def symmetric real
        np.random.seed(1234)
        data = np.random.rand(9, 9)
        data = np.dot(data.conj(), data.T)
        self.cases.append(Case("rand-sym-pd", data))
        # note: minres fails for single precision
        self.cases.append(Case("rand-sym-pd-F", data.astype('f'),
                               skip=[minres]))

        # Random complex-valued
        rng = np.random.RandomState(1234)
        data = rng.rand(4, 4) + 1j * rng.rand(4, 4)
        skip_cmplx = posdef_solvers + sym_solvers + real_solvers
        self.cases.append(Case("rand-cmplx", data, skip=skip_cmplx))
        self.cases.append(Case("rand-cmplx-F", data.astype('F'),
                               skip=skip_cmplx))

        # Random hermitian complex-valued
        rng = np.random.RandomState(1234)
        data = rng.rand(4, 4) + 1j * rng.rand(4, 4)
        data = data + data.T.conj()
        self.cases.append(Case("rand-cmplx-herm", data,
                               skip=posdef_solvers + real_solvers))
        self.cases.append(Case("rand-cmplx-herm-F", data.astype('F'),
                               skip=posdef_solvers + real_solvers))

        # Random pos-def hermitian complex-valued
        rng = np.random.RandomState(1234)
        data = rng.rand(9, 9) + 1j * rng.rand(9, 9)
        data = np.dot(data.conj(), data.T)
        self.cases.append(Case(
            "rand-cmplx-sym-pd", data, skip=real_solvers)
        )
        self.cases.append(Case("rand-cmplx-sym-pd-F", data.astype('F'),
                               skip=real_solvers))

        # Non-symmetric and Positive Definite
        #
        # cgs, qmr, bicg and tfqmr fail to converge on this one
        #   -- algorithmic limitation apparently
        data = ones((2, 10))
        data[0, :] = 2
        data[1, :] = -1
        A = dia_array((data, [0, -1]), shape=(10, 10)).tocsr()
        self.cases.append(Case("nonsymposdef", A,
                               skip=sym_solvers + [cgs, qmr, bicg, tfqmr]))
        self.cases.append(Case("nonsymposdef-F", A.astype('F'),
                               skip=sym_solvers + [cgs, qmr, bicg, tfqmr]))

        # Symmetric, non-pd, hitting cgs/bicg/bicgstab/qmr/tfqmr breakdown
        A = np.array([[0, 0, 0, 0, 0, 1, -1, -0, -0, -0, -0],
                      [0, 0, 0, 0, 0, 2, -0, -1, -0, -0, -0],
                      [0, 0, 0, 0, 0, 2, -0, -0, -1, -0, -0],
                      [0, 0, 0, 0, 0, 2, -0, -0, -0, -1, -0],
                      [0, 0, 0, 0, 0, 1, -0, -0, -0, -0, -1],
                      [1, 2, 2, 2, 1, 0, -0, -0, -0, -0, -0],
                      [-1, 0, 0, 0, 0, 0, -1, -0, -0, -0, -0],
                      [0, -1, 0, 0, 0, 0, -0, -1, -0, -0, -0],
                      [0, 0, -1, 0, 0, 0, -0, -0, -1, -0, -0],
                      [0, 0, 0, -1, 0, 0, -0, -0, -0, -1, -0],
                      [0, 0, 0, 0, -1, 0, -0, -0, -0, -0, -1]], dtype=float)
        b = np.array([0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0], dtype=float)
        assert (A == A.T).all()
        self.cases.append(Case("sym-nonpd", A, b,
                               skip=posdef_solvers,
                               nonconvergence=[cgs, bicg, bicgstab, qmr, tfqmr]
                               )
                          )

    def generate_tests(self):
        # generate test cases with skips applied
        tests = []
        for case in self.cases:
            for solver in _SOLVERS:
                if (solver in case.skip):
                    continue
                if solver in case.nonconvergence:
                    tests += [SingleTest(case.A, case.b, solver, case.name,
                                         convergence=False)]
                else:
                    tests += [
                        SingleTest(case.A, case.b, solver, case.name)
                    ]
        return tests


cases = IterativeParams().generate_tests()


@pytest.fixture(
    params=[pytest.param(c, marks=make_xp_pytest_marks(c.solver)) for c in cases],
    ids=[x.name for x in cases],
    scope="module"
)
def case(request):
    """
    Fixture for all cases in IterativeParams
    """
    return request.param


@pytest.mark.parametrize("batch_A", [()])
@pytest.mark.parametrize("batch_b", [()])
def test_maxiter(case, xp, batch_A, batch_b):
    case = xp_case(case, xp, batch_A, batch_b, rng=38)
    if not case.convergence:
        pytest.skip("Solver - Breakdown case, see gh-8829")
    A = case.A
    rtol = 1e-12

    b = case.b
    x0 = 0 * b

    residuals = []

    def callback(x):
        if x.ndim == 0:
            residuals.append(xp_vector_norm(b - case.A * x))
        else:
            Ax = xp.squeeze(case.A @ x[..., xp.newaxis], axis=-1)
            residuals.append(xp_vector_norm(b - Ax, axis=-1))

    if case.solver == gmres:
        with pytest.warns(DeprecationWarning, match=CB_TYPE_FILTER):
            x, info = case.solver(A, b, x0=x0, rtol=rtol, maxiter=1, callback=callback)
    else:
        x, info = case.solver(A, b, x0=x0, rtol=rtol, maxiter=1, callback=callback)

    assert len(residuals) == 1
    assert info == 1


@pytest.mark.parametrize("batch_A", [()])
@pytest.mark.parametrize("batch_b", [()])
def test_convergence(case, xp, batch_A, batch_b):
    if (case.solver is tfqmr) and ("poisson2d-F" in case.name):
        pytest.skip("Struggles to converge with single precision on some platforms")
    case = xp_case(case, xp, batch_A, batch_b, rng=38)
    A = case.A

    if A.dtype in (xp.float64, xp.complex128):
        rtol = 1e-8
    else:
        rtol = 1e-2

    b = case.b
    x0 = 0 * b

    x, info = case.solver(A, b, x0=x0, rtol=rtol)
    xp_assert_equal(x0, 0 * b)  # ensure that x0 is not overwritten

    if case.convergence:
        assert info == 0
        _assert_success(A=A, x=x, b=b, xp=xp, rtol=rtol)
    else:
        assert info != 0
        _assert_success(A=A, x=x, b=b, xp=xp, rtol=1, less_equal=True)


@pytest.mark.parametrize("batch_A", [()])
@pytest.mark.parametrize("batch_b", [()])
def test_precond_dummy(case, xp, batch_A, batch_b):
    dtype = case.A.dtype
    case = xp_case(case, xp, batch_A, batch_b)
    if (case.solver is cgs) and ("pd-F" in case.name):
        pytest.skip("Struggles to converge with single precision")
    if (case.solver is tfqmr) and ("poisson2d-F" in case.name):
        pytest.skip("Hits divide-by-zero with single precision")
    if not case.convergence:
        pytest.skip("Solver - Breakdown case, see gh-8829")

    rtol = 1e-8 if np.finfo(dtype).eps < 1e-8 else 1.2e-3

    A = case.A
    
    # NOTE: the following was previously uncommented as dead code --
    # was the intention to set `A = dia_array(...)`?

    # _, M, N = A.shape
    # # Ensure the diagonal elements of A are non-zero before calculating
    # # 1.0 / xp.linalg.diagonal(A)
    # diagOfA = xp.linalg.diagonal(A) if not is_numpy(xp) else A.diagonal()
    # if xp.count_nonzero(diagOfA) == diagOfA.shape[0]:
    #     dia_array(([1.0 / diagOfA], [0]), shape=(M, N))

    b = case.b
    x0 = 0 * b

    precond = IdentityOperator(shape=A.shape)

    if case.solver is qmr:
        x, info = case.solver(A, b, M1=precond, M2=precond, x0=x0, rtol=rtol)
    else:
        x, info = case.solver(A, b, M=precond, x0=x0, rtol=rtol)

    assert info == 0
    _assert_success(A=A, x=x, b=b, xp=np, rtol=rtol)

    A = aslinearoperator(A)

    x, info = case.solver(A, b, x0=x0, rtol=rtol)
    assert info == 0
    _assert_success(A=A, x=x, b=b, xp=np, rtol=rtol)


@pytest.mark.parametrize("batch_A", [()])
@pytest.mark.parametrize("batch_b", [()])
@pytest.mark.fail_slow(10)
def test_precond_inverse(case, xp, batch_A, batch_b):
    if case.casename not in ('poisson1d', 'poisson2d'):
        pytest.skip("specific to poisson1d and poisson2d cases")
    if case.solver is qmr:
        pytest.skip("skipped for qmr")
    
    case = xp_case(case, xp, batch_A, batch_b, rng=38)
    rtol = 1e-8

    def inverse(b, which=None):
        """inverse preconditioner"""
        A = case.A
        if is_numpy(xp) and not isinstance(A, np.ndarray):
            A = A.toarray()
        return xp.linalg.solve(A, b[..., np.newaxis])

    def rinverse(b, which=None):
        """inverse preconditioner"""
        A = case.A
        if is_numpy(xp) and not isinstance(A, np.ndarray):
            A = A.toarray()
        return xp.linalg.solve(A.T, b[..., np.newaxis])

    matvec_count = [0]

    def matvec(b):
        matvec_count[0] += 1
        return case.A @ b[..., np.newaxis]

    def rmatvec(b):
        matvec_count[0] += 1
        return case.A.T @ b[..., np.newaxis]

    b = case.b
    x0 = 0 * b

    A = LinearOperator(case.A.shape, matvec, rmatvec=rmatvec)
    precond = LinearOperator(case.A.shape, inverse, rmatvec=rinverse)

    # Solve with preconditioner
    matvec_count = [0]
    x, info = case.solver(A, b, M=precond, x0=x0, rtol=rtol)

    assert info == 0
    _assert_success(A=case.A, x=x, b=b, xp=xp, rtol=rtol)

    # Solution should be nearly instant
    assert matvec_count[0] <= 3


@pytest.mark.skip_xp_backends("dask.array", reason="https://github.com/dask/dask/issues/11711")
@pytest.mark.parametrize("batch_A", [()])
@pytest.mark.parametrize("batch_b", [()])
def test_atol(solver, xp, batch_A, batch_b):
    # TODO: minres / tfqmr. It didn't historically use absolute tolerances, so
    # fixing it is less urgent.
    if solver in (minres, tfqmr):
        pytest.skip("TODO: Add atol to minres/tfqmr")

    # Historically this is tested as below, all pass but for some reason
    # gcrotmk is over-sensitive to difference between random.seed/rng.random
    # Hence tol lower bound is changed from -10 to -9
    # np.random.seed(1234)
    # A = np.random.rand(10, 10)
    # A = A @ A.T + 10 * np.eye(10)
    # b = 1e3*np.random.rand(10)

    rng = np.random.default_rng(168441431005389)
    A = rng.uniform(size=(*batch_A, 10, 10))
    A = A @ A.mT + 10*np.eye(10)
    b = 1e3 * rng.uniform(size=(*batch_b, 10))

    dtype = xpx.default_dtype(xp)
    A = xp.asarray(A, dtype=dtype) 
    b = xp.asarray(b, dtype=dtype)

    tols = np.r_[0, np.logspace(-9, 2, 7), np.inf]

    # Check effect of badly scaled preconditioners
    M0 = rng.standard_normal(size=(*batch_A, 10, 10))
    M0 = xp.asarray(M0)
    M0 = M0 @ M0.mT
    Ms = [None, 1e-6 * M0, 1e6 * M0]

    for M, rtol, atol in itertools.product(Ms, tols, tols):
        if rtol == 0 and atol == 0:
            continue

        if solver is qmr:
            if M is not None:
                M = aslinearoperator(M)
                M2 = IdentityOperator(shape=(*batch_A, 10, 10))
            else:
                M2 = None
            x, info = solver(A, b, M1=M, M2=M2, rtol=rtol, atol=atol)
        else:
            x, info = solver(A, b, M=M, rtol=rtol, atol=atol)

        assert info == 0
        _assert_success(A=A, x=x, b=b, xp=xp, rtol=rtol, atol=atol)


@pytest.mark.skip_xp_backends("dask.array", reason="https://github.com/dask/dask/issues/11711")
@pytest.mark.parametrize("batch_A", [()])
@pytest.mark.parametrize("batch_b", [()])
def test_zero_rhs(solver, xp, batch_A, batch_b):
    rng = np.random.default_rng(1684414984100503)
    dtype = xpx.default_dtype(xp) 
    A = xp.asarray(rng.random(size=(*batch_A, 10, 10)), dtype=dtype)
    A = A @ A.mT + 10 * xp.eye(10)

    b = xp.zeros((*batch_b, 10), dtype=dtype)
    tols = np.r_[np.logspace(-10, 2, 7)]

    expected = xp.broadcast_to(b, (*np.broadcast_shapes(batch_A, batch_b), 10))
    for tol in tols:
        tol = float(tol)
        x, info = solver(A, b, rtol=tol)
        assert info == 0
        xp_assert_close(x, expected, atol=1e-15)

        x, info = solver(A, b, rtol=tol, x0=xp.ones((*batch_b, 10)))
        assert info == 0
        xp_assert_close(x, expected, atol=tol)

        if solver is not minres:
            x, info = solver(A, b, rtol=tol, atol=0.0, x0=xp.ones((*batch_b, 10)))
            if info == 0:
                xp_assert_close(x, expected)

            x, info = solver(A, b, rtol=tol, atol=tol)
            assert info == 0
            xp_assert_close(x, expected, atol=1e-300)

            x, info = solver(A, b, rtol=tol, atol=0.0)
            assert info == 0
            xp_assert_close(x, expected, atol=1e-300)


# unbatched test
@pytest.mark.xfail(reason="see gh-18697")
def test_maxiter_worsening(solver, xp):
    if solver not in (gmres, lgmres, qmr):
        # these were skipped from the very beginning, see gh-9201; gh-14160
        pytest.skip("Solver breakdown case")
    # Check error does not grow (boundlessly) with increasing maxiter.
    # This can occur due to the solvers hitting close to breakdown,
    # which they should detect and halt as necessary.
    # cf. gh-9100
    if (solver is lgmres and
            platform.machine() not in ['x86_64' 'x86', 'aarch64', 'arm64']):
        # see gh-17839
        pytest.xfail(reason="fails on at least ppc64le, ppc64 and riscv64")

    # Singular matrix, rhs numerically not in range
    A = np.array([[-0.1112795288033378, 0, 0, 0.16127952880333685],
                  [0, -0.13627952880333782 + 6.283185307179586j, 0, 0],
                  [0, 0, -0.13627952880333782 - 6.283185307179586j, 0],
                  [0.1112795288033368, 0j, 0j, -0.16127952880333785]])
    v = np.ones(4)
    dtype = xpx.default_dtype(xp)
    A, v = (xp.asarray(arr, dtype=dtype) for arr in [A, v])
    best_error = np.inf

    # Unable to match the Fortran code tolerance levels with this example
    # Original tolerance values

    # slack_tol = 7 if platform.machine() == 'aarch64' else 5
    slack_tol = 9

    rtol = 1e-8
    for maxiter in range(1, 20):
        x, info = solver(A, v, maxiter=maxiter, rtol=rtol, atol=0.0)
        if info == 0:
            _assert_success(A=A, x=x, b=v, xp=xp, rtol=rtol)

        # Check with slack
        error = xp_vector_norm(A @ x - v)
        best_error = xp.min(best_error, error)
        assert error <= slack_tol * best_error


def _setup_random_system(xp, batch_A, batch_b):
    rng = np.random.default_rng(1685363802304750)
    n = 10
    A = rng.random(size=(*batch_A, n, n))
    A = A @ A.mT
    b = rng.random((*batch_b, n))
    x0 = rng.random((*batch_b, n))
    dtype = xpx.default_dtype(xp)
    A, b, x0 = (xp.asarray(arr, dtype=dtype) for arr in [A, b, x0])
    return A, b, x0


@pytest.mark.parametrize("batch_A", [()])
@pytest.mark.parametrize("batch_b", [()])
def test_x0_working(solver, xp, batch_A, batch_b):
    # Easy problem
    A, b, x0 = _setup_random_system(xp, batch_A, batch_b)

    if solver is minres:
        kw = dict(rtol=1e-6)
    else:
        kw = dict(atol=0.0, rtol=1e-6)

    x, info = solver(A, b, **kw)
    assert info == 0

    _assert_success(A=A, x=x, b=b, xp=xp, rtol=1e-6)

    x, info = solver(A, b, x0=x0, **kw)
    assert info == 0
    _assert_success(A=A, x=x, b=b, xp=xp, rtol=1e-5)


@pytest.mark.parametrize("batch_A", [()])
@pytest.mark.parametrize("batch_b", [()])
def test_x0_equals_Mb(case, xp, batch_A, batch_b):
    dtype = case.A.dtype
    case = xp_case(case, xp, batch_A, batch_b, rng=38)
    if (case.solver is cgs) and ("pd-F" in case.name):
        pytest.skip("Struggles to converge with single precision")
    if (case.solver is bicgstab) and (case.name == 'nonsymposdef-bicgstab'):
        pytest.skip("Solver fails due to numerical noise "
                    "on some architectures (see gh-15533).")
    if case.solver is tfqmr:
        pytest.skip("Solver does not support x0='Mb'")

    A = case.A
    b = case.b
    x0 = 'Mb'
    rtol = 1e-8 if np.finfo(dtype).eps < 1e-8 else 1.5e-3
    x, info = case.solver(A, b, x0=x0, rtol=rtol)

    assert x0 == 'Mb'  # ensure that x0 is not overwritten
    assert info == 0
    _assert_success(A=A, x=x, b=b, xp=xp, rtol=rtol)


@pytest.mark.parametrize("batch_A", [()])
@pytest.mark.parametrize("batch_b", [()])
def test_x0_solves_problem_exactly(solver, xp, batch_A, batch_b):
    # See gh-19948
    dtype = xpx.default_dtype(xp)
    mat = xp.tile(xp.eye(2, dtype=dtype), (*batch_A, 1, 1))
    rhs = xp.tile(xp.asarray([-1., -1.], dtype=dtype), (*batch_b, 1))

    sol, info = solver(mat, rhs, x0=rhs)
    expected = xp.broadcast_to(rhs, (*np.broadcast_shapes(batch_A, batch_b), 2))
    xp_assert_close(sol, expected)
    assert info == 0


@pytest.mark.parametrize("batch_A", [()])
@pytest.mark.parametrize("batch_b", [()])
def test_show(case, capsys, xp, batch_A, batch_b):
    if case.solver != tfqmr:
        pytest.skip("tfqmr specific test")
    if "poisson2d-F" in case.name:
        pytest.skip("Hits divide-by-zero with single precision")
    if "pd-F" in case.name:
        pytest.skip("Struggles to converge with single precision on some platforms")
    case = xp_case(case, xp, batch_A, batch_b)
    def cb(x):
        pass

    ctx = (
        np.errstate(all='ignore')
        if case.casename == "nonsymposdef"
        else nullcontext()
    )
    with ctx:
        x, info = tfqmr(case.A, case.b, callback=cb, show=True)

    out, err = capsys.readouterr()

    if case.casename == "sym-nonpd":
        # no logs for some reason
        exp = ""
    elif case.casename in ("nonsymposdef", "nonsymposdef-F"):
        # Asymmetric and Positive Definite
        exp = "TFQMR: Linear solve not converged due to reach MAXIT iterations"
    else:  # all other cases
        exp = "TFQMR: Linear solve converged due to reach TOL iterations"

    assert out.startswith(exp)
    assert err == ""


@pytest.mark.parametrize("batch_A", [()])
@pytest.mark.parametrize("batch_b", [()])
def test_positional_error(solver, xp, batch_A, batch_b):
    A, b, x0 = _setup_random_system(xp, batch_A, batch_b)
    with pytest.raises(TypeError):
        solver(A, b, x0, 1e-5)


@pytest.mark.parametrize("batch_A", [()])
@pytest.mark.parametrize("batch_b", [()])
@pytest.mark.parametrize("atol", ["legacy", None, -1])
def test_invalid_atol(solver, atol, xp, batch_A, batch_b):
    if solver == minres:
        pytest.skip("minres has no `atol` argument")
    A, b, x0 = _setup_random_system(xp, batch_A, batch_b)
    with pytest.raises(ValueError):
        solver(A, b, x0, atol=atol)


class TestQMR:
    @pytest.mark.filterwarnings('ignore::scipy.sparse.SparseEfficiencyWarning')
    def test_leftright_precond(self):
        """Check that QMR works with left and right preconditioners"""

        from scipy.sparse.linalg._dsolve import splu
        from scipy.sparse.linalg._interface import LinearOperator

        n = 100

        dat = ones(n)
        A = dia_array(([-2 * dat, 4 * dat, -dat], [-1, 0, 1]), shape=(n, n))
        b = arange(n, dtype='d')

        L = dia_array(([-dat / 2, dat], [-1, 0]), shape=(n, n))
        U = dia_array(([4 * dat, -dat], [0, 1]), shape=(n, n))
        L_solver = splu(L)
        U_solver = splu(U)

        def L_solve(b):
            return L_solver.solve(b)

        def U_solve(b):
            return U_solver.solve(b)

        def LT_solve(b):
            return L_solver.solve(b, 'T')

        def UT_solve(b):
            return U_solver.solve(b, 'T')

        M1 = LinearOperator((n, n), matvec=L_solve, rmatvec=LT_solve)
        M2 = LinearOperator((n, n), matvec=U_solve, rmatvec=UT_solve)

        rtol = 1e-8
        x, info = qmr(A, b, rtol=rtol, maxiter=15, M1=M1, M2=M2)

        assert info == 0
        assert xp_vector_norm(A @ x - b) <= rtol * xp_vector_norm(b)


class TestGMRES:
    def test_basic(self):
        A = np.vander(np.arange(10) + 1)[:, ::-1]
        b = np.zeros(10)
        b[0] = 1

        x_gm, err = gmres(A, b, restart=5, maxiter=1)

        assert_allclose(x_gm[0], 0.359, rtol=1e-2)

    @pytest.mark.filterwarnings(f"ignore:{CB_TYPE_FILTER}:DeprecationWarning")
    def test_callback(self):

        def store_residual(r, rvec):
            rvec[rvec.nonzero()[0].max() + 1] = r

        # Define, A,b
        A = csr_array(array([[-2, 1, 0, 0, 0, 0],
                             [1, -2, 1, 0, 0, 0],
                             [0, 1, -2, 1, 0, 0],
                             [0, 0, 1, -2, 1, 0],
                             [0, 0, 0, 1, -2, 1],
                             [0, 0, 0, 0, 1, -2]]))
        b = ones((A.shape[0],))
        maxiter = 1
        rvec = zeros(maxiter + 1)
        rvec[0] = 1.0

        def callback(r):
            return store_residual(r, rvec)

        x, flag = gmres(A, b, x0=zeros(A.shape[0]), rtol=1e-16,
                        maxiter=maxiter, callback=callback)

        # Expected output from SciPy 1.0.0
        assert_allclose(rvec, array([1.0, 0.81649658092772603]), rtol=1e-10)

        # Test preconditioned callback
        M = 1e-3 * np.eye(A.shape[0])
        rvec = zeros(maxiter + 1)
        rvec[0] = 1.0
        x, flag = gmres(A, b, M=M, rtol=1e-16, maxiter=maxiter,
                        callback=callback)

        # Expected output from SciPy 1.0.0
        # (callback has preconditioned residual!)
        assert_allclose(rvec, array([1.0, 1e-3 * 0.81649658092772603]),
                        rtol=1e-10)

    def test_abi(self):
        # Check we don't segfault on gmres with complex argument
        A = eye(2)
        b = ones(2)
        r_x, r_info = gmres(A, b)
        r_x = r_x.astype(complex)
        x, info = gmres(A.astype(complex), b.astype(complex))

        assert iscomplexobj(x)
        assert_allclose(r_x, x)
        assert r_info == info

    @pytest.mark.fail_slow(10)
    def test_atol_legacy(self):

        A = eye(2)
        b = ones(2)
        x, info = gmres(A, b, rtol=1e-5)
        assert xp_vector_norm(A @ x - b) <= 1e-5 * xp_vector_norm(b)
        assert_allclose(x, b, atol=0.0, rtol=1e-8)

        rndm = np.random.RandomState(12345)
        A = rndm.rand(30, 30)
        b = 1e-6 * ones(30)
        x, info = gmres(A, b, rtol=1e-7, restart=20)
        assert xp_vector_norm(A @ x - b) > 1e-7

        A = eye(2)
        b = 1e-10 * ones(2)
        x, info = gmres(A, b, rtol=1e-8, atol=0.0)
        assert xp_vector_norm(A @ x - b) <= 1e-8 * xp_vector_norm(b)

    def test_defective_precond_breakdown(self):
        # Breakdown due to defective preconditioner
        M = np.eye(3)
        M[2, 2] = 0

        b = np.array([0, 1, 1])
        x = np.array([1, 0, 0])
        A = np.diag([2, 3, 4])

        x, info = gmres(A, b, x0=x, M=M, rtol=1e-15, atol=0.0)

        # Should not return nans, nor terminate with false success
        assert not np.isnan(x).any()
        if info == 0:
            assert xp_vector_norm(A @ x - b) <= 1e-15 * xp_vector_norm(b)

        # The solution should be OK outside null space of M
        assert_allclose(M @ (A @ x), M @ b)

    def test_defective_matrix_breakdown(self):
        # Breakdown due to defective matrix
        A = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]])
        b = np.array([1, 0, 1])
        rtol = 1e-8
        x, info = gmres(A, b, rtol=rtol, atol=0)

        # Should not return nans, nor terminate with false success
        assert not np.isnan(x).any()
        if info == 0:
            assert xp_vector_norm(A @ x - b) <= rtol * xp_vector_norm(b)

        # The solution should be OK outside null space of A
        assert_allclose(A @ (A @ x), A @ b)

    @pytest.mark.filterwarnings(f"ignore:{CB_TYPE_FILTER}:DeprecationWarning")
    def test_callback_type(self):
        # The legacy callback type changes meaning of 'maxiter'
        np.random.seed(1)
        A = np.random.rand(20, 20)
        b = np.random.rand(20)

        cb_count = [0]

        def pr_norm_cb(r):
            cb_count[0] += 1
            assert isinstance(r, float)

        def x_cb(x):
            cb_count[0] += 1
            assert isinstance(x, np.ndarray)

        # 2 iterations is not enough to solve the problem
        cb_count = [0]
        x, info = gmres(A, b, rtol=1e-6, atol=0.0, callback=pr_norm_cb,
                        maxiter=2, restart=50)
        assert info == 2
        assert cb_count[0] == 2

        # With `callback_type` specified, no warning should be raised
        cb_count = [0]
        x, info = gmres(A, b, rtol=1e-6, atol=0.0, callback=pr_norm_cb,
                        maxiter=2, restart=50, callback_type='legacy')
        assert info == 2
        assert cb_count[0] == 2

        # 2 restart cycles is enough to solve the problem
        cb_count = [0]
        x, info = gmres(A, b, rtol=1e-6, atol=0.0, callback=pr_norm_cb,
                        maxiter=2, restart=50, callback_type='pr_norm')
        assert info == 0
        assert cb_count[0] > 2

        # 2 restart cycles is enough to solve the problem
        cb_count = [0]
        x, info = gmres(A, b, rtol=1e-6, atol=0.0, callback=x_cb, maxiter=2,
                        restart=50, callback_type='x')
        assert info == 0
        assert cb_count[0] == 1

    def test_callback_x_monotonic(self):
        # Check that callback_type='x' gives monotonic norm decrease
        rng = np.random.RandomState(1)
        A = rng.rand(20, 20) + np.eye(20)
        b = rng.rand(20)

        prev_r = [np.inf]
        count = [0]

        def x_cb(x):
            r = xp_vector_norm(A @ x - b)
            assert r <= prev_r[0]
            prev_r[0] = r
            count[0] += 1

        x, info = gmres(A, b, rtol=1e-6, atol=0.0, callback=x_cb, maxiter=20,
                        restart=10, callback_type='x')
        assert info == 20
        assert count[0] == 20


def test_nD(solver, xp):
    """Check that >2-D operators are rejected cleanly."""
    def id(x):
        return x
    A = LinearOperator(shape=(2, 2, 2), matvec=id, dtype=xp.float64, xp=xp)
    b = xp.ones((2, 2))
    with pytest.raises(ValueError, match="expected 2-D"):
        solver(A, b)



from numpy.testing import assert_allclose
from scipy.linalg import cython_lapack as cython_lapack
from scipy.linalg import lapack


class TestLamch:

    def test_slamch(self):
        lapack_slamch = lapack.get_lapack_funcs(
            'lamch', dtype='float32', ilp64='preferred'
        )
        for c in [b'e', b's', b'b', b'p', b'n', b'r', b'm', b'u', b'l', b'o']:
            assert_allclose(cython_lapack._test_slamch(c),
                            lapack_slamch(c))

    def test_dlamch(self):
        lapack_dlamch = lapack.get_lapack_funcs(
            'lamch', dtype='float64', ilp64='preferred'
        )

        for c in [b'e', b's', b'b', b'p', b'n', b'r', b'm', b'u', b'l', b'o']:
            assert_allclose(cython_lapack._test_dlamch(c),
                            lapack_dlamch(c))

    def test_complex_ladiv(self):
        cx = .5 + 1.j
        cy = .875 + 2.j
        assert_allclose(cython_lapack._test_zladiv(cy, cx), 1.95+0.1j)
        assert_allclose(cython_lapack._test_cladiv(cy, cx), 1.95+0.1j)


class TestBlasInt:
    """Test for blas_int/blas_bint types used in cython_blas/cython_lapack.

    Note similar test is in test_blas.py: both cython_blas.pyx and cython_lapack.pyx
    define the types, so we smoke test that `blas_int` and `blas_bint` types
    can be cimported from both `cython_blas` and `cython_lapack` extensions.
    """

    def test_blas_int_size(self):
        """Verify blas_int and blas_bint sizes matches the build configuration."""
        from scipy.__config__ import CONFIG
        size = cython_lapack._blas_int_size()
        cython_blas_ilp64 = CONFIG['Build Dependencies']['blas']['cython blas ilp64']
        if cython_blas_ilp64:
            assert size == 8, f"ILP64 build but blas_int is {size} bytes"
        else:
            assert size == 4, f"LP64 build but blas_int is {size} bytes"

        # also verify blas_bint size
        # bint_sizes is a 2-tuple, (sizeof(blas_bint), sizeof(bint))
        bint_sizes = cython_lapack._blas_bint_bint_sizes()
        if cython_blas_ilp64:
            assert bint_sizes == (8, 4)
        else:
            assert bint_sizes == (4, 4)



'''
This module provides Powell's COBYLA algorithm.

Translated from Zaikun Zhang's modern-Fortran reference implementation in PRIMA.

Dedicated to late Professor M. J. D. Powell FRS (1936--2015).

Python translation by Nickolai Belakovski.

N.B.:

1. The modern-Fortran reference implementation in PRIMA contains bug fixes and improvements over the
original Fortran 77 implementation by Powell. Consequently, the PRIMA implementation behaves differently
from the original Fortran 77 implementation by Powell. Therefore, it is important to point out that
you are using PRIMA rather than the original solvers if you want your results to be reproducible.

2. Compared to Powell's Fortran 77 implementation, the modern-Fortran implementation and hence any
faithful translation like this one generally produce better solutions with fewer function evaluations,
making them preferable for applications with expensive function evaluations. However, if function
evaluations are not the dominant cost in your application, the Fortran 77 solvers are likely to be
faster, as they are more efficient in terms of memory usage and flops thanks to the careful and
ingenious (but unmaintained and unmaintainable) implementation by Powell.

See the PRIMA documentation (www.libprima.net) for more information.
'''

from ..common.evaluate import evaluate, moderatex, moderatef, moderatec
from ..common.consts import (EPS, RHOBEG_DEFAULT, RHOEND_DEFAULT, CTOL_DEFAULT,
                                   CWEIGHT_DEFAULT, FTARGET_DEFAULT, IPRINT_DEFAULT,
                                   MAXFUN_DIM_DEFAULT, DEBUGGING, BOUNDMAX,
                                   ETA1_DEFAULT, ETA2_DEFAULT, GAMMA1_DEFAULT,
                                   GAMMA2_DEFAULT)
from ..common.preproc import preproc
from ..common.present import present
from ..common.linalg import matprod
from .cobylb import cobylb
import numpy as np
from dataclasses import dataclass
from copy import copy


@dataclass
class COBYLAResult:
    x: np.ndarray
    f: float
    constr: np.ndarray
    cstrv: float
    nf: int
    xhist: np.ndarray | None
    fhist: np.ndarray | None
    chist: np.ndarray | None
    conhist: np.ndarray | None
    info: int


def cobyla(calcfc, m_nlcon, x, Aineq=None, bineq=None, Aeq=None, beq=None,
           xl=None, xu=None, f0=None, nlconstr0=None, rhobeg=None, rhoend=None,
           ftarget=FTARGET_DEFAULT, ctol=CTOL_DEFAULT, cweight=CWEIGHT_DEFAULT,
           maxfun=None, iprint=IPRINT_DEFAULT, eta1=None, eta2=None,
           gamma1=GAMMA1_DEFAULT, gamma2=GAMMA2_DEFAULT, maxhist=None, maxfilt=2000,
           callback=None):
    """
    Among all the arguments, only CALCFC, M_NLCON, and X are obligatory. The others are
    OPTIONAL and you can neglect them unless you are familiar with the algorithm. Any
    unspecified optional input will take the default value detailed below. For
    instance, we may invoke the solver as follows.

    # First define CALCFC, M_NLCON, and X, and then do the following.
    result = cobyla(calcfc, m_nlcon, x)

    or

    # First define CALCFC, M_NLCON, X, Aineq, and Bineq, and then do the following.
    result = cobyla(calcfc, m_nlcon, x, Aineq=Aineq, bineq=bineq, rhobeg=1.0e0,
        rhoend=1.0e-6)

    ####################################################################################
    # IMPORTANT NOTICE: The user must set M_NLCON correctly to the number of nonlinear
    # constraints, namely the size of NLCONSTR introduced below. Set it to 0 if there
    # is no nonlinear constraint.
    ####################################################################################

    See examples/cobyla/cobyla_example.py for a concrete example.

    A detailed introduction to the arguments is as follows.

    ####################################################################################
    # INPUTS
    ####################################################################################

    CALCFC
      Input, function.
      f, nlconstr = CALCFC(X) should evaluate the objective function and nonlinear
      constraints at the given vector X; it should return a tuple consisting of the
      objective function value and the nonlinear constraint value. It must be provided
      by the user, and its definition must conform to the following interface:
      #-------------------------------------------------------------------------#
       def calcfc(x):
           f = 0.0
           nlconstr = np.zeros(m_nlcon)
           return f, nlconstr
      #-------------------------------------------------------------------------#

    M_NLCON
      Input, scalar.
      M_NLCON must be set to the number of nonlinear constraints, namely the size of
      NLCONSTR(X).
      N.B.:
      1. Why don't we define M_NLCON as optional and default it to 0 when it is absent?
      This is because we need to allocate memory for CONSTR_LOC using M_NLCON. To
      ensure that the size of CONSTR_LOC is correct, we require the user to specify
      M_NLCON explicitly.

    X
      Input, vector.
      As an input, X should be an N-dimensional vector that contains the starting
      point, N being the dimension of the problem.

    Aineq, Bineq
      Input, matrix of size [Mineq, N] and vector of size Mineq unless they are both
      empty, default: None and None.
      Aineq and Bineq represent the linear inequality constraints: Aineq*X <= Bineq.

    Aeq, Beq
      Input, matrix of size [Meq, N] and vector of size Meq unless they are both
      empty, default: None and None.
      Aeq and Beq represent the linear equality constraints: Aeq*X = Beq.

    XL, XU
      Input, vectors of size N unless they are both None, default: None and None.
      XL is the lower bound for X. If XL is None, X has no
      lower bound. Any entry of XL that is NaN or below -BOUNDMAX will be taken as
      -BOUNDMAX, which effectively means there is no lower bound for the corresponding
      entry of X. The value of BOUNDMAX is 0.25*HUGE(X), which is about 8.6E37 for
      single precision and 4.5E307 for double precision. XU is similar.

    F0
      Input, scalar.
      F0, if present, should be set to the objective function value of the starting X.

    NLCONSTR0
      Input, vector.
      NLCONSTR0, if present, should be set to the nonlinear constraint value at the
      starting X; in addition, SIZE(NLCONSTR0) must be M_NLCON, or the solver will
      abort.

    RHOBEG, RHOEND
      Inputs, scalars, default: RHOBEG = 1, RHOEND = 10^-6. RHOBEG and RHOEND must be
      set to the initial and final values of a trust-region radius, both being positive
      and RHOEND <= RHOBEG. Typically RHOBEG should be about one tenth of the greatest
      expected change to a variable, and RHOEND should indicate the accuracy that is
      required in the final values of the variables.

    FTARGET
      Input, scalar, default: -Inf.
      FTARGET is the target function value. The algorithm will terminate when a
      feasible point with a function value <= FTARGET is found.

    CTOL
      Input, scalar, default: sqrt(machine epsilon).
      CTOL is the tolerance of constraint violation. X is considered feasible if
      CSTRV(X) <= CTOL.
      N.B.:
        1. CTOL is absolute, not relative.
        2. CTOL is used only when selecting the returned X. It does not affect the
           iterations of the algorithm.

    CWEIGHT
      Input, scalar, default: CWEIGHT_DFT defined in common/consts.py.
      CWEIGHT is the weight that the constraint violation takes in the selection of the
      returned X.

    MAXFUN
      Input, integer scalar, default: MAXFUN_DIM_DFT*N with MAXFUN_DIM_DFT defined in
      common/consts.py. MAXFUN is the maximal number of calls of CALCFC.

    IPRINT
      Input, integer scalar, default: 0.
      The value of IPRINT should be set to 0, 1, -1, 2, -2, 3, or -3, which controls
      how much information will be printed during the computation:
      0: there will be no printing;
      1: a message will be printed to the screen at the return, showing the best vector
         of variables found and its objective function value;
      2: in addition to 1, each new value of RHO is printed to the screen, with the
         best vector of variables so far and its objective function value; each new
         value of CPEN is also printed;
      3: in addition to 2, each function evaluation with its variables will be printed
         to the screen; -1, -2, -3: the same information as 1, 2, 3 will be printed,
         not to the screen but to a file named COBYLA_output.txt; the file will be
         created if it does not exist; the new output will be appended to the end of
         this file if it already exists.
      Note that IPRINT = +/-3 can be costly in terms of time and/or space.

    ETA1, ETA2, GAMMA1, GAMMA2
      Input, scalars, default: ETA1 = 0.1, ETA2 = 0.7, GAMMA1 = 0.5, and GAMMA2 = 2.
      ETA1, ETA2, GAMMA1, and GAMMA2 are parameters in the updating scheme of the
      trust-region radius detailed in the subroutine TRRAD in trustregion.py. Roughly
      speaking, the trust-region radius is contracted by a factor of GAMMA1 when the
      reduction ratio is below ETA1, and enlarged by a factor of GAMMA2 when the
      reduction ratio is above ETA2. It is required that 0 < ETA1 <= ETA2 < 1 and
      0 < GAMMA1 < 1 < GAMMA2. Normally, ETA1 <= 0.25. It is NOT advised to set
      ETA1 >= 0.5.

    MAXFILT
      Input, scalar.
      MAXFILT is a nonnegative integer indicating the maximal length of the filter used
      for selecting the returned solution; default: MAXFILT_DFT (a value lower than
      MIN_MAXFILT is not recommended);
      see common/consts.py for the definitions of MAXFILT_DFT and MIN_MAXFILT.

    CALLBACK
      Input, function to report progress and optionally request termination.


    ####################################################################################
    # OUTPUTS
    ####################################################################################

    The output is a single data structure, COBYLAResult, with the following fields:

    X
      Output, vector.
      As an output, X will be set to an approximate minimizer.

    F
      Output, scalar.
      F will be set to the objective function value of X at exit.

    CONSTR
      Output, vector.
      CONSTR will be set to the constraint value of X at exit.

    CSTRV
      Output, scalar.
      CSTRV will be set to the constraint violation of X at exit, i.e.,
      max([0, XL - X, X - XU, Aineq*X - Bineq, ABS(Aeq*X -Beq), NLCONSTR(X)]).

    NF
      Output, scalar.
      NF will be set to the number of calls of CALCFC at exit.

    XHIST, FHIST, CHIST, CONHIST, MAXHIST
      XHIST: Output, rank 2 array;
      FHIST: Output, rank 1 array;
      CHIST: Output, rank 1 array;
      CONHIST: Output, rank 2 array;
      MAXHIST: Input, scalar, default: MAXFUN
      XHIST, if present, will output the history of iterates; FHIST, if present, will
      output the history function values; CHIST, if present, will output the history of
      constraint violations; CONHIST, if present, will output the history of constraint
      values; MAXHIST should be a nonnegative integer, and XHIST/FHIST/CHIST/CONHIST
      will output only the history of the last MAXHIST iterations.
      Therefore, MAXHIST= 0 means XHIST/FHIST/CONHIST/CHIST will output
      nothing, while setting MAXHIST = MAXFUN requests XHIST/FHIST/CHIST/CONHIST to
      output all the history. If XHIST is present, its size at exit will be
      (N, min(NF, MAXHIST)); if FHIST is present, its size at exit will be
      min(NF, MAXHIST); if CHIST is present, its size at exit will be min(NF, MAXHIST);
      if CONHIST is present, its size at exit will be (M, min(NF, MAXHIST)).

      IMPORTANT NOTICE:
      Setting MAXHIST to a large value can be costly in terms of memory for large
      problems.
      MAXHIST will be reset to a smaller value if the memory needed exceeds MAXHISTMEM
      defined in common/consts.py
      Use *HIST with caution!!! (N.B.: the algorithm is NOT designed for large
      problems).

    INFO
      Output, scalar.
      INFO is the exit flag. It will be set to one of the following values defined in
      common/infos.py:
      SMALL_TR_RADIUS: the lower bound for the trust region radius is reached;
      FTARGET_ACHIEVED: the target function value is reached;
      MAXFUN_REACHED: the objective function has been evaluated MAXFUN times;
      MAXTR_REACHED: the trust region iteration has been performed MAXTR times (MAXTR = 2*MAXFUN);
      NAN_INF_X: NaN or Inf occurs in X;
      DAMAGING_ROUNDING: rounding errors are becoming damaging.
      #--------------------------------------------------------------------------#
      The following case(s) should NEVER occur unless there is a bug.
      NAN_INF_F: the objective function returns NaN or +Inf;
      NAN_INF_MODEL: NaN or Inf occurs in the model;
      TRSUBP_FAILED: a trust region step failed to reduce the model
      #--------------------------------------------------------------------------#
    """

    # Local variables
    solver = "COBYLA"
    srname = "COBYLA"

    # Sizes
    mineq = len(bineq) if present(bineq) else 0
    meq = len(beq) if present(beq) else 0
    mxl = sum(xl > -BOUNDMAX) if present(xl) else 0
    mxu = sum(xu < BOUNDMAX) if present(xu) else 0
    mmm = mxu + mxl + 2*meq + mineq + m_nlcon
    num_vars = len(x)

    # Preconditions
    if DEBUGGING:
        assert m_nlcon >= 0, f'{srname} M_NLCON >= 0'
        assert num_vars >= 1, f'{srname} N >= 1'

        assert present(Aineq) == present(bineq), \
            f'{srname} Aineq and Bineq are both present or both absent'
        if (present(Aineq)):
            assert Aineq.shape == (mineq, num_vars), f'{srname} SIZE(Aineq) == [Mineq, N]'

        assert present(Aeq) == present(beq), \
            f'{srname} Aeq and Beq are both present or both absent'
        if (present(Aeq)):
            assert Aeq.shape == (meq, num_vars), f'{srname} SIZE(Aeq) == [Meq, N]'

        if (present(xl)):
            assert len(xl) == num_vars, f'{srname} SIZE(XL) == N'
        if (present(xu)):
            assert len(xu) == num_vars, f'{srname} SIZE(XU) == N'


        # N.B.: If NLCONSTR0 is present, then F0 must be present, and we assume that
        # F(X0) = F0 even if F0 is NaN; if NLCONSTR0 is absent, then F0 must be either
        # absent or NaN, both of which will be interpreted as F(X0) is not provided.
        if present(nlconstr0):
            assert present(f0), f'{srname} If NLCONSTR0 is present, then F0 is present'
        if present(f0):
            assert np.isnan(f0) or present(nlconstr0), \
                f'{srname} If F0 is present and not NaN, then NLCONSTR0 is present'



    # Exit if the size of NLCONSTR0 is inconsistent with M_NLCON.
    if present(nlconstr0):
        assert np.size(nlconstr0) == m_nlcon

    # Read the inputs.

    if xl is not None:
      xl = copy(xl)
      xl[np.isnan(xl)] = -BOUNDMAX
      xl[xl < -BOUNDMAX] = -BOUNDMAX

    if xu is not None:
      xu = copy(xu)
      xu[np.isnan(xu)] = BOUNDMAX
      xu[xu > BOUNDMAX] = BOUNDMAX

    # Wrap the linear and bound constraints into a single constraint: AMAT@X <= BVEC.
    amat, bvec = get_lincon(Aeq, Aineq, beq, bineq, xl, xu)

    # Create constraint vector
    constr = np.zeros(mmm)

    # Set [F_LOC, CONSTR_LOC] to [F(X0), CONSTR(X0)] after evaluating the latter if
    # needed. In this way, COBYLB only needs one interface.
    # N.B.: Due to the preconditions above, there are two possibilities for F0 and
    # NLCONSTR0.
    # If NLCONSTR0 is present, then F0 must be present, and we assume that F(X0) = F0
    # even if F0 is NaN.
    # If NLCONSTR0 is absent, then F0 must be either absent or NaN, both of which will
    # be interpreted as F(X0) is not provided and we have to evaluate F(X0) and
    # NLCONSTR(X0) now.
    if (present(f0) and present(nlconstr0) and all(np.isfinite(x))):
        f = moderatef(f0)
        if amat is not None:
          constr[:mmm - m_nlcon] = moderatec(matprod(amat, x) - bvec)
        constr[mmm - m_nlcon:] = moderatec(nlconstr0)
    else:
        x = moderatex(x)
        f, constr = evaluate(calcfc, x, m_nlcon, amat, bvec)
        constr[:mmm - m_nlcon] = moderatec(constr[:mmm - m_nlcon])
        # N.B.: Do NOT call FMSG, SAVEHIST, or SAVEFILT for the function/constraint evaluation at X0.
        # They will be called during the initialization, which will read the function/constraint at X0.
    cstrv = max(np.append(0, constr))


    # If RHOBEG is present, use it; otherwise, RHOBEG takes the default value for
    # RHOBEG, taking the value of RHOEND into account. Note that RHOEND is considered
    # only if it is present and it is VALID (i.e., finite and positive). The other
    # inputs are read similarly.
    if present(rhobeg):
        rhobeg = rhobeg
    elif present(rhoend) and np.isfinite(rhoend) and rhoend > 0:
        rhobeg = max(10 * rhoend, RHOBEG_DEFAULT)
    else:
        rhobeg = RHOBEG_DEFAULT

    if present(rhoend):
        rhoend = rhoend
    elif rhobeg > 0:
        rhoend = max(EPS, min(RHOEND_DEFAULT/RHOBEG_DEFAULT * rhobeg, RHOEND_DEFAULT))
    else:
        rhoend = RHOEND_DEFAULT

    maxfun = maxfun if present(maxfun) else MAXFUN_DIM_DEFAULT * num_vars

    if present(eta1):
        eta1 = eta1
    elif present(eta2) and 0 < eta2 < 1:
        eta1 = max(EPS, eta2 / 7)
    else:
        eta1 = ETA1_DEFAULT

    if present(eta2):
        eta2 = eta2
    elif 0 < eta1 < 1:
        eta2 = (eta1 + 2) / 3
    else:
        eta2 = ETA2_DEFAULT

    maxhist = (
        maxhist
        if present(maxhist)
        else max(maxfun, num_vars + 2, MAXFUN_DIM_DEFAULT * num_vars)
    )

    # Preprocess the inputs in case some of them are invalid. It does nothing if all
    # inputs are valid.
    (
        iprint,
        maxfun,
        maxhist,
        ftarget,
        rhobeg,
        rhoend,
        npt,  # Unused in COBYLA
        maxfilt,
        ctol,
        cweight,
        eta1,
        eta2,
        gamma1,
        gamma2,
        _x0,  # Unused in COBYLA
    ) = preproc(
        solver,
        num_vars,
        iprint,
        maxfun,
        maxhist,
        ftarget,
        rhobeg,
        rhoend,
        num_constraints=mmm,
        maxfilt=maxfilt,
        ctol=ctol,
        cweight=cweight,
        eta1=eta1,
        eta2=eta2,
        gamma1=gamma1,
        gamma2=gamma2,
        is_constrained=(mmm > 0),
    )

    # Further revise MAXHIST according to MAXHISTMEM, and allocate memory for the history.
    # In MATLAB/Python/Julia/R implementation, we should simply set MAXHIST = MAXFUN and initialize
    # CHIST = NaN(1, MAXFUN), CONHIST = NaN(M, MAXFUN), FHIST = NaN(1, MAXFUN), XHIST = NaN(N, MAXFUN)
    # if they are requested; replace MAXFUN with 0 for the history that is not requested.
    # prehist(maxhist, num_vars, present(xhist), xhist_loc, present(fhist), fhist_loc, &
    #     & present(chist), chist_loc, m, present(conhist), conhist_loc)

    # call cobylb, which performs the real calculations
    x, f, constr, cstrv, nf, xhist, fhist, chist, conhist, info = cobylb(
        calcfc,
        iprint,
        maxfilt,
        maxfun,
        amat,
        bvec,
        ctol,
        cweight,
        eta1,
        eta2,
        ftarget,
        gamma1,
        gamma2,
        rhobeg,
        rhoend,
        constr,
        f,
        x,
        maxhist,
        callback
    )

    return COBYLAResult(x, f, constr, cstrv, nf, xhist, fhist, chist, conhist, info)


def get_lincon(Aeq=None, Aineq=None, beq=None, bineq=None, xl=None, xu=None):
    """
    This subroutine wraps the linear and bound constraints into a single constraint:
        AMAT*X <= BVEC.

    N.B.:

    LINCOA normalizes the linear constraints so that each constraint has a gradient
    of norm 1. However, COBYLA does not do this.
    """

    # Sizes
    if Aeq is not None:
        num_vars = Aeq.shape[1]
    elif Aineq is not None:
        num_vars = Aineq.shape[1]
    elif xl is not None:
        num_vars = len(xl)
    elif xu is not None:
        num_vars = len(xu)
    else:
        return None, None

    # Preconditions
    if DEBUGGING:
        assert Aineq is None or Aineq.shape == (len(bineq), num_vars)
        assert Aeq is None or Aeq.shape == (len(beq), num_vars)
        assert (xl is None or xu is None) or len(xl) == len(xu) == num_vars

    #====================#
    # Calculation starts #
    #====================#

    # Define the indices of the nontrivial bound constraints.
    ixl = np.where(xl > -BOUNDMAX)[0] if xl is not None else None
    ixu = np.where(xu < BOUNDMAX)[0] if xu is not None else None

    # Wrap the linear constraints.
    # The bound constraint XL <= X <= XU is handled as two constraints:
    # -X <= -XL, X <= XU.
    # The equality constraint Aeq*X = Beq is handled as two constraints:
    # -Aeq*X <= -Beq, Aeq*X <= Beq.
    # N.B.:
    # 1. The treatment of the equality constraints is naive. One may choose to
    #    eliminate them instead.
    idmat = np.eye(num_vars)
    amat = np.vstack([
        -idmat[ixl, :] if ixl is not None else np.empty((0, num_vars)),
        idmat[ixu, :] if ixu is not None else np.empty((0, num_vars)),
        -Aeq if Aeq is not None else np.empty((0, num_vars)),
        Aeq if Aeq is not None else np.empty((0, num_vars)),
        Aineq if Aineq is not None else np.empty((0, num_vars))
    ])
    bvec = np.hstack([
        -xl[ixl] if ixl is not None else np.empty(0),
        xu[ixu] if ixu is not None else np.empty(0),
        -beq if beq is not None else np.empty(0),
        beq if beq is not None else np.empty(0),
        bineq if bineq is not None else np.empty(0)
    ])

    amat = amat if amat.shape[0] > 0 else None
    bvec = bvec if bvec.shape[0] > 0 else None

    #==================#
    # Calculation ends #
    #==================#

    # Postconditions
    if DEBUGGING:
        assert (amat is None and bvec is None) or amat.shape == (len(bvec), num_vars)

    return amat, bvec



# This file is not meant for public use and will be removed in SciPy v2.0.0.
# Use the `scipy.ndimage` namespace for importing the functions
# included below.

from scipy._lib.deprecation import _sub_module_deprecation


__all__ = [  # noqa: F822
    'iterate_structure', 'generate_binary_structure',
    'binary_erosion', 'binary_dilation', 'binary_opening',
    'binary_closing', 'binary_hit_or_miss', 'binary_propagation',
    'binary_fill_holes', 'grey_erosion', 'grey_dilation',
    'grey_opening', 'grey_closing', 'morphological_gradient',
    'morphological_laplace', 'white_tophat', 'black_tophat',
    'distance_transform_bf', 'distance_transform_cdt',
    'distance_transform_edt'
]


def __dir__():
    return __all__


def __getattr__(name):
    return _sub_module_deprecation(sub_package='ndimage', module='morphology',
                                   private_modules=['_morphology'], all=__all__,
                                   attribute=name)
